/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.gui;

import java.util.List;
import java.util.Observable;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.xml.namespace.QName;

import kgrid.raems.stubs.RAEMSService.RAEMSPortType;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.NotifyCallback;
import org.globus.wsrf.WSNConstants;
import org.globus.wsrf.core.notification.ResourcePropertyValueChangeNotificationElementType;
import org.oasis.wsn.Subscribe;
import org.oasis.wsn.TopicExpressionType;
import org.oasis.wsrf.properties.ResourcePropertyValueChangeNotificationType;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 */
public class NotificationHandler extends Observable implements NotifyCallback {
	public static final String NS = "http://grid.deis.unical.it/kgrid/RAEMSService";
	public static final String STATE_NOTIF = "#STATE:",
			ERROR_NOTIF = "#ERROR:", SERVICE_NOTIF = "#SERVICE:";
	private ConcurrentLinkedQueue<String> notifications = new ConcurrentLinkedQueue<String>();
	private NotificationFeeder feeder;

	public void subscribeToRAEMSModelTopic(RAEMSPortType raems)
			throws Exception {

		// 1: set up the notification consumer
		// (the listener will be set up as a notification consumer)
		NotificationConsumerManager consumer = null;

		QName RAEMS_RP_MODEL = new QName(NS, "RAEMSModel");

		// 2: set the 'listener' as a 'notification consumer'
		consumer = NotificationConsumerManager.getInstance();
		consumer.startListening();
		EndpointReferenceType consumerEPR = consumer
				.createNotificationConsumer(this);

		// 3: create (and opportunetely set) the remote subscribe request
		// Create the request to the remote Subscribe() call
		Subscribe request = new Subscribe();
		// Must the notification be delivered using the Notify
		// operation?
		request.setUseNotify(Boolean.TRUE);
		// Indicate what the consumer client's EPR is
		request.setConsumerReference(consumerEPR);

		// 4: create the topic to be subscribed to
		// The TopicExpression specifies what topic we want to
		// subscribe to
		TopicExpressionType topicExpression = new TopicExpressionType();
		topicExpression.setDialect(WSNConstants.SIMPLE_TOPIC_DIALECT);
		topicExpression.setValue(RAEMS_RP_MODEL);
		request.setTopicExpression(topicExpression);

		// 6: send the subscribe request to the notification producer (KDS)
		// IMPORTANT: the consumer DASListener is subscribed to the 'KDS_Model'
		// topic
		raems.subscribe(request);
	}// subscribe

	public void deliver(List topicPath, EndpointReferenceType producer,
			Object message) {
		synchronized (notifications) {
			ResourcePropertyValueChangeNotificationElementType notif_elem;
			ResourcePropertyValueChangeNotificationType notif;

			notif_elem = (ResourcePropertyValueChangeNotificationElementType) message;
			notif = notif_elem.getResourcePropertyValueChangeNotification();

			if (notif != null) {
				for (MessageElement me : notif.getNewValue().get_any()) {
					notifications.add(me.getValue());
				}
			}
		}
	}

	public Notification parse(String msg) {
		Notification n = null;
		if (msg == null)
			return null;
		else if (msg.startsWith(ERROR_NOTIF)) {
			String name = msg.substring(ERROR_NOTIF.length(), msg.indexOf('$'));
			String errMsg = msg.substring(msg.indexOf('$') + 1, msg.length());
			n = new Notification(name);
			n.setError(errMsg);
		} else if (msg.startsWith(STATE_NOTIF)) {
			String name = msg.substring(STATE_NOTIF.length(), msg.indexOf('$'));
			String status = msg.substring(msg.indexOf('$') + 1, msg.length());
			n = new Notification(name);
			n.setStatus(getTaskStateEnumeration(status));
		} else if (msg.startsWith(SERVICE_NOTIF)) {
			String result = msg.substring(SERVICE_NOTIF.length());
			n = new Notification("");
			n.setFinish(result);
		}
		return n;
	}

	private TaskState getTaskStateEnumeration(String status) {
		if (status.equalsIgnoreCase("CREATED")) {
			return TaskState.CREATED;
		} else if (status.equalsIgnoreCase("ACTIVE")) {
			return TaskState.RUNNING;
		} else if (status.equalsIgnoreCase("DONE")) {
			return TaskState.DONE;
		} else if (status.equalsIgnoreCase("FAILED")) {
			return TaskState.FAILED;
		} else if (status.equalsIgnoreCase("DESTROYED")) {
			return TaskState.DESTROYED_UNKNOWN;
		}
		return TaskState.NOT_CONNECTED;
	}

	public void startFeeding() {
		feeder = new NotificationFeeder();
		feeder.setFeeding(true);
		feeder.start();
	}

	public void stopFeeding() {
		feeder.setFeeding(false);
	}

	public class NotificationFeeder extends Thread {
		private boolean feed = false;

		public void run() {
			while (isFeeding()) {

				String msg = notifications.poll();
				if (msg != null) {
					setChanged();
					notifyObservers(msg);
					clearChanged();
				}
				try {
					Thread.sleep(80);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}

		public synchronized void setFeeding(boolean feed) {
			this.feed = feed;
		}

		public synchronized boolean isFeeding() {
			return feed;
		}
	}
}
