/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.gui;

import java.awt.BorderLayout;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Observable;
import java.util.Observer;

import javax.swing.Box;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;

import kgrid.raems.stubs.RAEMSService.ExecutionParameters;
import kgrid.raems.stubs.RAEMSService.ExecutionResults;
import kgrid.raems.stubs.RAEMSService.RAEMSPortType;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.xml.serialize.DOMSerializer;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.oasis.wsrf.lifetime.Destroy;
import org.w3c.dom.Document;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 */
public class RAEMSClient extends JFrame implements Observer,
        PropertyObserver<Task> {

    // import com.sun.xml.internal.bind.v2.runtime.XMLSerializer;
    private static final long serialVersionUID = -2444007318281302967L;
    private ExecutionPlanVisualizationViewer abstract_viewer, concrete_viewer;
    private JTabbedPane                      graphTabs;
    private Box                              mainBox, secondaryBox;
    private PropertyVisualization            property;
    private ConsoleArea                      console;
    private Toolbar                          toolbar;
    private Document                         abstractEP;
    private RAEMSThread                      runner = null;

    @Override
    public void update(Observable o, Object arg) {
        if (o instanceof NotificationHandler) {
            String msg = (String) arg;
            Notification n = ((NotificationHandler) o).parse(msg);
            if (n != null) {
                if (n.isErrorNotification()) {
                    console.appendLine("Task " + n.getTaskName()
                            + " has stopped running. Caused By:"
                            + n.getErrorMessage());
                }
                if (n.isStatusNotification()) {
                    abstract_viewer.getGraph().setTaskStatus(n.getTaskName(),
                            n.getStatus());
                    concrete_viewer.getGraph().setTaskStatus(n.getTaskName(),
                            n.getStatus());
                }
                if (n.isFinishNotification()) {
                    console.appendLine("Service has finished with result: "
                            + n.getFinishMessage());
                    if (runner != null)
                        try {
                            console.append("Destroying RAEMS Resource...");
                            runner.destroyRaems();
                            console.appendLine("Done");
                        } catch (Exception e) {
                            displayError(e);
                        }
                }
            } else {
                console.appendLine(msg);
            }

        }
    }

    @Override
    public void displayProperties(Collection<Task> propsHolders) {
        Iterator<Task> it = propsHolders.iterator();
        String total = "";
        while (it.hasNext()) {
            Task t = it.next();
            total += "Task:" + t.getName() + "\nCurrent Status:"
                    + t.getStatus() + "\nStatus History:\n" + t.getHistory();
            total += "\n----------------\n";
        }
        property.displayProperty(total);

    }

    @Override
    public void displayProperty(Task propHolder) {
        property.displayProperty("Task:" + propHolder.getName()
                + "\nCurrent Status:" + propHolder.getStatus()
                + "\nStatus History:\n" + propHolder.getHistory());

    }

    public void displayError(Throwable t) {
        if (console != null) {
            console.append(t.getLocalizedMessage());
        }
        JOptionPane.showMessageDialog(this, t.getLocalizedMessage(),
                "An error occurred", JOptionPane.ERROR_MESSAGE);
        t.printStackTrace();
    }

    public void init() {
        try {
            toolbar = new Toolbar(this);
            toolbar.init();
            this.add(toolbar, BorderLayout.NORTH);

            mainBox = Box.createVerticalBox();
            secondaryBox = Box.createHorizontalBox();
            abstract_viewer = new ExecutionPlanVisualizationViewer();
            abstract_viewer
                    .addGraphMousePlugin(new PropertyGraphMousePlugin<Task, String>(
                            MouseEvent.BUTTON1_MASK, this));
            concrete_viewer = new ExecutionPlanVisualizationViewer();
            concrete_viewer
                    .addGraphMousePlugin(new PropertyGraphMousePlugin<Task, String>(
                            MouseEvent.BUTTON1_MASK, this));
            property = new PropertyVisualization();
            property.enableProperties(false);
            graphTabs = new JTabbedPane();
            graphTabs.addTab("abstract", abstract_viewer.getComponent());
            graphTabs.addTab("concrete", concrete_viewer.getComponent());
            graphTabs.setEnabledAt(0, false);
            graphTabs.setEnabledAt(1, false);
            // graphTabs.setSelectedComponent(concrete_viewer.getComponent());
            secondaryBox.add(graphTabs);
            // secondaryBox.add(abstract_viewer.getComponent());
            secondaryBox.add(property);
            console = new ConsoleArea();
            console.init();
            mainBox.add(secondaryBox);
            mainBox.add(console);
            this.add(mainBox, BorderLayout.CENTER);
            this.pack();
            this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            this.setVisible(true);
            console.appendLine("Initialization completed.");
            console.append("Precaching Images...");
            IconCache.buildIconCache(this);
            console.appendLine("Done");
        } catch (Throwable t) {
            displayError(t);
        }
    }

    public void openInsertIPDialog() {
        InputParamDialog dialog = new InputParamDialog(this);
        try {
            dialog
                    .init(new String[] { "RAEMS ip:" },
                            new String[] { InetAddress.getLocalHost()
                                    .getHostAddress() });
        } catch (UnknownHostException e) {
            displayError(e);
        }
        dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
        // dialog.setModalityType(Dialog.ModalityType.APPLICATION_MODAL);
        dialog.setLocationRelativeTo(this);
        // dialog.
        dialog.setVisible(true);
        dialog.dispose();
        if (dialog.hasInput()) {
            connectAndRun(dialog.getInput().get("RAEMS ip:"));
        }
    }

    public void openLoadInputDialog() {
        JFileChooser dialog = new JFileChooser(new File("./abstractmodel"));
        int returnVal = dialog.showOpenDialog(this);
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            loadGraph(dialog.getSelectedFile().getAbsolutePath());
        }

    }

    private void loadGraph(String eppath) {
        try {
            abstractEP = RAEMSUtils.loadFromFile(eppath);
            displayGraph(abstractEP);
            JOptionPane.showMessageDialog(this, "Loading Complete");
        } catch (Exception e) {
            displayError(e);
        }
    }

    private void connectAndRun(String ip) {
        try {
            console.appendLine("Starting WS Client...");
            runner = new RAEMSThread(ip, abstractEP, this);
            runner.start();
            console.appendLine("WS Client Started.");
        } catch (Exception e) {
            displayError(e);
        }
    }

    public void displayConcreteExecutionPlan(Document concreteExecutionPlan)
            throws IOException, InterruptedException {
        ExecutionPlanGraph graph = new ExecutionPlanGraph();
        graph.init(concreteExecutionPlan);
        concrete_viewer.setGraph(graph);
        graphTabs.setEnabledAt(1, true);
        ViewConsoleArea vca = new ViewConsoleArea();
        vca.setEditable(false);
        printDocument(concreteExecutionPlan, vca.getPrintStream());
        graphTabs.addTab("concreteDoc", new JScrollPane(vca));
    }

    private void printDocument(Document doc, PrintStream stream)
            throws IOException {
        DOMSerializer ser = new XMLSerializer(stream, new OutputFormat(doc,
                "UTF-8", true));
        ser.serialize(doc);

    }// printDocument

    private void displayGraph(Document abstractEP) throws IOException,
            InterruptedException {
        console.appendLine("Loading Execution Plan...");
        ExecutionPlanGraph graph = new ExecutionPlanGraph();
        graph.init(abstractEP);
        abstract_viewer.setGraph(graph);
        console.appendLine("Loading completed.");
        graphTabs.setEnabledAt(0, true);
        // abstract_viewer.fireStateChanged();
        // graphTabs.removeTabAt(0);
        // graphTabs.addTab("abstract", abstract_viewer.getComponent());
        // displayConcreteExecutionPlan(abstractEP);
    }

    public static void main(String[] args) {
        RAEMSClient client = new RAEMSClient();
        client.init();
        // client.setVisible(true);
    }

    private class RAEMSThread extends Thread {

        private String              ip;
        private Document            ep;
        private RAEMSClient         observer;
        private RAEMSPortType       raems;               ;
        private ExecutionParameters params;
        private NotificationHandler notificationHandler;

        public RAEMSThread(String ip, Document abstractEP, RAEMSClient obs)
                throws Exception {
            this.ip = ip;
            this.ep = abstractEP;
            this.observer = obs;
            observer.console.appendLine("Delegating User Credentials...");
            EndpointReferenceType cred = RAEMSUtils.delegateCredential(ip);
            observer.console.appendLine("Credentials Delegated");
            params = new ExecutionParameters(cred, ep);
            raems = RAEMSUtils.getRAEMSPortType(ip);
            notificationHandler = new NotificationHandler();
            notificationHandler.addObserver(observer);
            observer.console
                    .appendLine("Subscribing to RAEMS Notification Topic...");
            notificationHandler.subscribeToRAEMSModelTopic(raems);
            observer.console.appendLine("Client Subscribed Done");
        }

        public void run() {
            try {
                observer.console
                        .appendLine("Invoking RAEMS Port Type.Waiting for Concrete Execution Plan...");
                ExecutionResults results = raems.manageKExecution(params);
                observer.displayConcreteExecutionPlan(results
                        .getConcreteExecutionPlan());
                observer.console
                        .appendLine("Concrete Execution Plan delivered and displayed");
                observer.console.appendLine("Listening to Notification...");
                notificationHandler.startFeeding();
                // saveDocument(results.getConcreteExecutionPlan(),"./results/cep"
                // );

            } catch (Throwable e) {
                observer.displayError(e);
            }
        }

        private void saveDocument(Document doc, String fileName)
                throws IOException {

            PrintWriter pw = new PrintWriter(new FileWriter(fileName + ".xml"));
            DOMSerializer ser = new XMLSerializer(pw, new OutputFormat(doc,
                    "UTF-8", true));
            ser.serialize(doc);
        }

        public void destroyRaems() throws Exception {
            raems.destroy(new Destroy());
        }
    }
}
