/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.net.UnknownHostException;
import java.rmi.RemoteException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import kgrid.kds.stubs.KDSService.ArrayOfString;
import kgrid.raems.gui.NotificationHandler;
import kgrid.raems.service.impl.exception.CreateEndpointReferenceException;
import kgrid.raems.service.impl.exception.EmptyResultSetException;
import kgrid.raems.service.impl.exception.SystemPermissionException;
import kgrid.raems.service.impl.execution.ExecutionHandler;
import kgrid.raems.service.impl.scheduler.Mapper;
import kgrid.raems.service.impl.scheduler.ResourceMapper;
import kgrid.raems.service.impl.scheduler.Scheduler;
import kgrid.raems.service.impl.scheduler.SimpleMapper;
import kgrid.raems.service.impl.scheduler.SimpleResourceMapper;
import kgrid.raems.service.impl.scheduler.SimpleRoundRobinResourceMapper;
import kgrid.raems.service.impl.utils.KDSClient;
import kgrid.raems.service.impl.utils.Query;
import kgrid.raems.service.impl.utils.Security;
import kgrid.raems.service.impl.utils.Session;
import kgrid.raems.service.impl.utils.XmlUtils;
import kgrid.raems.service.impl.workflow.Deserializer;
import kgrid.raems.service.impl.workflow.Job;
import kgrid.raems.service.impl.workflow.PluginDescription;
import kgrid.raems.service.impl.workflow.plugin.DataMiningSoftwareQuerable;
import kgrid.raems.service.impl.workflow.plugin.DatasetQuerable;
import kgrid.raems.stubs.RAEMSService.ExecutionParameters;
import kgrid.raems.stubs.RAEMSService.ExecutionResults;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.globus.delegation.DelegationException;
import org.globus.wsrf.ResourceContext;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class RAEMSService implements Consts {

	/**
	 * Reference to the session folder. A new session is created at every
	 * invocation. Session includes the delegated client credentials, a session
	 * ID, and the current resource
	 */
	private Session session = null;

	// ======================= P O R T T Y P E =======================

	public ArrayOfString getQueryFromXML(String xml) {
		Document document;
		Query query = null;
		ArrayOfString aos = new ArrayOfString(new String[] {});
		try {
			document = xmlToDocument(xml);
			Utils.printDocToScreen(document);
			NodeList nList = document.getChildNodes();
			for (int i = 0; i < nList.getLength(); i++) {
				String nodeName = nList.item(i).getNodeName();
				System.out.println("nodeName = " + nodeName);
				if (nodeName.equals("DataMiningSoftware")) {
					DataMiningSoftwareQuerable dmSoftwareQuery = new DataMiningSoftwareQuerable();
					dmSoftwareQuery.update(document);
					query = dmSoftwareQuery.getQuery();
					break;
				} else if (nodeName.equals("Dataset")) {
					DatasetQuerable datasetQuery = new DatasetQuerable();
					datasetQuery.update(document);
					query = datasetQuery.getQuery();
					break;
				}
			}
			aos = query.getArrayOfString();
		} catch (Exception e) {
			e.printStackTrace();
		}
		System.out.println("Query translated:");
		for (String s : aos.getX()) {
			System.out.println(s);
		}
		return aos;
	}

	/**
	 * The only public method in this service. Manage the execution of an
	 * Abstract Execution Plan through building a concrete Execution Plan then
	 * scheduling and executing it on some grid nodes.
	 * 
	 * @param parameters
	 * @return
	 * @throws RemoteException
	 */
	public ExecutionResults manageKExecution(ExecutionParameters parameters)
			throws RemoteException {
		try {
			createSession(parameters.getEprCredential());
			Job concreteExecutionPlan = createConcreteExecutionPlan(parameters
					.getExecutionPlanDoc());
			session.setResourceModel("Concrete Execution Plan successfully built");
			Document jobXML = saveConcreteExecutionPlan(concreteExecutionPlan);
			// let's exit from the manageKExecution and leave the concrete
			// Execution Plan to the RAEMSThread
			Thread execThread = new RAEMSThread(this, concreteExecutionPlan);
			ExecutionResults ret = new ExecutionResults();
			ret.setConcreteExecutionPlan(jobXML);
			execThread.start();
			return ret;
		} catch (Exception e) {
			e.printStackTrace();
			if (e instanceof EmptyResultSetException) {
				String error = NotificationHandler.ERROR_NOTIF
						+ "one or more resource have no mappings. Can't execute Execution Plan";
				session.setResourceModel(error);
				Utils.print(error);
			} else {
				Utils.print(e.getMessage());
			}
			throw new RemoteException("An Error Occurred", e);
		}
	}

	// ======================= Utility Methods =======================

	private Document xmlToDocument(String xml) throws Exception {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		Document doc = builder.parse(new ByteArrayInputStream(xml.getBytes()));
		doc.normalize();
		return doc;
	}

	/**
	 * Saves the concrete Execution Plan in the session
	 */
	Document saveConcreteExecutionPlan(Job concreteExecutionPlan)
			throws IOException, ParserConfigurationException {
		File cepf = new File(SESSION_PATH + session.getSessionID()
				+ "/concreteExecutionPlan");
		if (cepf.exists())
			throw new RuntimeException(
					"A concrete execution plan already exist..");
		Document doc = concreteExecutionPlan.serializeConcreteExecutionPlan();
		XmlUtils.saveDocument(doc, SESSION_PATH + session.getSessionID()
				+ "/concreteExecutionPlan.xml");
		return doc;
	}

	/**
	 * Execute the Concrete Execution Plan. Should be invoked by RAEMSThread
	 */
	void executeConcreteExecutionPlan(Job concreteExecutionPlan)
			throws InterruptedException, RemoteException {
		ExecutionHandler executionHandler = new ExecutionHandler(
				concreteExecutionPlan, session);
		Utils.print("Executing concrete execution plan...");
		executionHandler.execute();
	}

	/**
	 * Constructs a concrete execution plan
	 */
	private Job createConcreteExecutionPlan(Document executionPlanDoc)
			throws Exception {
		PluginDescription pluginDescription = new PluginDescription(
				executionPlanDoc.getDocumentElement().getNodeName());
		Deserializer deserializer = new Deserializer(pluginDescription);
		Job abstractEP = (Job) deserializer.deserialize(executionPlanDoc);
		Job concreteEP = schedule(abstractEP);
		return concreteEP;
	}

	/**
	 * Starts scheduling the abstract Execution Plan. The scheduler both finds a
	 * match for abstract resources and detects and resolves hidden Data
	 * Transfers
	 * 
	 * @param abstractEP
	 * @return
	 * @throws Exception
	 */
	private Job schedule(Job abstractEP) throws Exception {
		KDSClient client = new KDSClient(session);
		ResourceMapper resourceMapper = new SimpleRoundRobinResourceMapper();
		Mapper mapper = new SimpleMapper(resourceMapper);
		Scheduler scheduler = new Scheduler(client, mapper);
		Job concreteEP = scheduler.schedule(abstractEP);
		return concreteEP;
	}

	/**
	 * Returns the RAEMS resource that this invocation of the service is using
	 */
	private RAEMSResource getResource() throws RemoteException {
		Object resource = null;
		resource = ResourceContext.getResourceContext().getResource();
		RAEMSResource raemsResource = (RAEMSResource) resource;
		return raemsResource;
	}

	/**
	 * Does some security stuff
	 */
	private EndpointReferenceType delegateCredentials(
			EndpointReferenceType eprCredential) throws DelegationException,
			CreateEndpointReferenceException, UnknownHostException {

		EndpointReferenceType newEprCredential = null;
		newEprCredential = Security.delegateCredential(Utils.HOST_IP,
				CONTAINER_PORT, eprCredential);

		return newEprCredential;
	}

	/**
	 * Creates a new session which consists of a directory in the KEPR
	 * repository named with the current time in milliseconds
	 */
	private void createSession(EndpointReferenceType eprCredential)
			throws SystemPermissionException {
		File f = null;
		try {
			String sessionID = "" + System.currentTimeMillis() + "/";
			session = new Session(getResource());
			session.setSessionID(sessionID);
			session.setClientCredentials(delegateCredentials(eprCredential));
			f = new File(SESSION_PATH + session.getSessionID());
			f.mkdirs();
		} catch (Exception e) {
			throw new SystemPermissionException("cannot create folder "
					+ session);
		}
		if (!f.setWritable(true, false))
			throw new SystemPermissionException(
					"cannot change folder permission: +w");
		if (!f.setReadable(true, false))
			throw new SystemPermissionException(
					"cannot change folder permission: +r");

	}

	Session getSession() {
		return session;
	}

}
