/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 * 
 * Portions of this file Copyright 1999-2005 University of Chicago Portions of
 * this file Copyright 1999-2005 The University of Southern California.
 * 
 * This file or a portion of this file is licensed under the terms of the Globus
 * Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html. If you redistribute this
 * file, with or without modifications, you must include this notice in the
 * file.
 */
package kgrid.raems.service.impl.execution;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;

import javax.xml.parsers.ParserConfigurationException;

import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.Utils;
import kgrid.raems.service.impl.exception.LocalExecutionError;
import kgrid.raems.service.impl.utils.XmlUtils;
import kgrid.raems.service.impl.workflow.Computation;
import kgrid.raems.service.impl.workflow.Task;
import kgrid.raems.service.impl.workflow.plugin.Arg;

import org.apache.axis.components.uuid.UUIDGen;
import org.apache.axis.components.uuid.UUIDGenFactory;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.globus.exec.client.GramJob;
import org.globus.exec.client.GramJobListener;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.utils.ManagedJobFactoryConstants;
import org.globus.exec.utils.client.ManagedJobFactoryClientHelper;
import org.globus.exec.utils.rsl.RSLHelper;
import org.globus.rft.generated.RFTOptionsType;
import org.globus.rft.generated.TransferRequestType;
import org.globus.rft.generated.TransferType;
import org.gridforum.jgss.ExtendedGSSManager;
import org.ietf.jgss.GSSCredential;
import org.xml.sax.SAXException;

/**
 * This class is a modified version of the official globusrun executable: It's a
 * 'Thread' It has no main method (not executable) Main function are left
 * unchanged, but it cooperates with the 'ExecutionHandler' class. Some unused
 * fields have been deleted. It doesn't have a 'ShutdownHook'
 * 
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class ComputationTask extends ExecutionThread implements
		GramJobListener, Consts {
	private static final long STATE_CHANGE_BASE_TIMEOUT_MILLIS = 60000;
	/**
	 * Job submission member variables.
	 */
	private GramJob job;
	private boolean jobCompleted = false; // completed
	// if
	// Done
	// or
	// Failed
	/**
	 * id used for job submission
	 */
	private String submissionID = null;

	/**
	 * indicates whether or not the job failed
	 */
	public ComputationTask(Task t, ExecutionHandler eh) {
		super(t, eh);
	}

	/**
	 * Poll for job status
	 * 
	 * @param job
	 * @throws Exception
	 */
	private void refreshJobStatus(GramJob job) throws Exception {
		job.refreshStatus();
	}

	/**
	 * Submits a previously created job description (RSL) to the gram service.
	 * The service is invoked using the specified factory EndpointReference.
	 * 
	 * @param factoryEndpoint
	 * @param rsl
	 * @throws Exception
	 */
	private void submitRSL(EndpointReferenceType factoryEndpoint,
			JobDescriptionType rsl) throws Exception {
		Utils.printJobCommandLine(rsl);
		if (rsl != null) {
			job = new GramJob(rsl);
		}
		job.setDelegationEnabled(true);
		processJob(job, factoryEndpoint);
	}

	/**
	 * Actually process the rsl using a GramJob object
	 * 
	 * @param job
	 * @param factoryEndpoint
	 * @throws Exception
	 */
	private void processJob(GramJob job, EndpointReferenceType factoryEndpoint)
			throws Exception {
		// ------Security------//
		ExtendedGSSManager manager = (ExtendedGSSManager) ExtendedGSSManager
				.getInstance();
		GSSCredential cred = manager
				.createCredential(GSSCredential.INITIATE_AND_ACCEPT);
		job.setCredentials(cred);
		// -----submissionID generation----//
		if (submissionID == null) {
			UUIDGen uuidgen = UUIDGenFactory.getUUIDGen();
			submissionID = "uuid:" + uuidgen.nextUUID();
		}

		// --------registering as a GramJobListener----------//
		job.addListener(this);
		job.submit(factoryEndpoint, false, true, submissionID);
		// wait till this job is completed or failed

		waitForJobCompletion(STATE_CHANGE_BASE_TIMEOUT_MILLIS);
		Utils.print("GRAM job completed");

		// no sync should be needed. every Thread should already been stopped

		// Destroying gram resource for task getTask().getName());

		destroyJob(job);
		state = State.DESTROYED;
		executionHandler.fireTaskResult(this);

		// Gram resource for task getTask().getName() destroyed
	}

	/**
	 * Since messaging is assumed to be unreliable (i.e. a notification could
	 * very well be lost), we implement policy of pulling the remote state when
	 * a given waited-for notification has not has been received after a
	 * timeout. Note: this could however have the side-effect of hiding bugs in
	 * the service-side notification implementation.
	 * 
	 * The base delay in parameter is doubled each time the wait times out
	 * (binary exponential backoff). When a state change notification is
	 * received, the time out delay is reset to the base value.
	 * 
	 * @param maxWaitPerStateNotificationMillis
	 *            long base timeout for each state transition before pulling the
	 *            state from the service
	 * @throws Exception
	 */
	private synchronized void waitForJobCompletion(
			long maxWaitPerStateNotificationMillis) throws Exception {

		long durationToWait = maxWaitPerStateNotificationMillis;
		long startTime;
		StateEnumeration oldState = job.getState();

		while (!jobCompleted) {
			/*
			 * Job not completed - waiting for state change. Timeout before
			 * pulling: " + durationToWait + " ms).");
			 */
			startTime = System.currentTimeMillis(); // (re)set start time
			wait(durationToWait); // wait for a state change notif

			// now let's determine what stopped the wait():

			StateEnumeration currentState = job.getState();
			// A) New job state change notification (good!)
			if (currentState != null && !currentState.equals(oldState)) {
				oldState = currentState; // wait for next state notif
				durationToWait = maxWaitPerStateNotificationMillis; // reset
			} else {
				long now = System.currentTimeMillis();
				long durationWaited = now - startTime;
				// B) Timeout when waiting for a notification (bad)
				if (durationWaited >= durationToWait) {
					Utils.print("Did not receive any new notification of "
							+ "job state change after a delay of "
							+ durationToWait + " ms.\nPulling job state.");
					// pull state from remote job and print the
					// state only if it is a new state
					refreshJobStatus(job);
					// Little issue: the printing of state
					// says "State Notification" even if it is not
					// really one...

					// binary exponential backoff
					durationToWait = 2 * durationToWait;
				}
				// C) Some other reason
				else {
					// wait but only for remainder of timeout duration
					durationToWait = durationToWait - durationWaited;
				}
			}
		}
	}

	/**
	 * Callback as a GramJobListener. Will not be called in batch mode. This
	 * method is invoked by the GRAM thread, and it will wake the main thread
	 * who is waiting for its termination.
	 */
	public synchronized void stateChanged(GramJob job) {
		State oldState = null;
		try {
			// manage state
			StateEnumeration jobState = job.getState();
			oldState = state;
			if (jobState.equals(StateEnumeration.Done)) {
				jobCompleted = true;
				state = State.DONE;
			} else if (jobState.equals(StateEnumeration.Failed)) {
				jobCompleted = true;
				state = State.FAILED;
			} else if (jobState.equals(StateEnumeration.Active)) {
				state = State.ACTIVE;
			}
		} catch (Exception e) {
			state = State.FAILED;
			throw new RuntimeException("Error while releasing GRAM Job", e);
		} finally {
			// notify the main thread that something has changed
			notifyAll();
			if (!state.equals(oldState)) {
				executionHandler.fireTaskResult(this);
			}
		}
	}

	/**
	 * destroys the job WSRF resource Precondition: job ! =null &&
	 * job.isRequested() && !job.isLocallyDestroyed() Gram job hanldes both gram
	 * resource and subscription resource destruction
	 */
	private void destroyJob(GramJob job) throws Exception {
		job.destroy();
	}

	/**
	 * Generate a job description (RSL) for its own computation
	 * 
	 * @return
	 * @throws SAXException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 */
	private JobDescriptionType getRsl() throws SAXException, IOException,
			ParserConfigurationException {
		Computation c = (Computation) task;
		JobDescriptionType descriptionType = new JobDescriptionType();
		descriptionType.setExecutable(c.getProgram().getInvocation());
		descriptionType.setStderr(c.getProgram().getStandardError());
		descriptionType.setStdin(c.getProgram().getStandardInput());
		descriptionType.setStdout(c.getProgram().getStandardOutput());
		descriptionType.setDirectory(KBR_PATH);

		// descriptionType.setJobCredentialEndpoint(jobCredential);
		/*
		 * if(!c.getProgram().getHostName().equals(InetAddress.getLocalHost().
		 * getHostName ()) &&
		 * !c.getProgram().getHostName().equals(InetAddress.getLocalHost()
		 * .getHostAddress())){ //
		 * descriptionType.setStagingCredentialEndpoint(stagingCredential);
		 * /*descriptionType
		 * .setFileStageOut(getStagingRequest(GlobalConfiguration
		 * .GSIFTP_PROTOCOL + c.getProgram().getHostName() + ":" + GSIPORT +
		 * KBR+c.getProgram().getStandardOutput (),GSIFTP_PROTOCOL +
		 * InetAddress.getLocalHost().getHostName() + ":" + GSIPORT +
		 * KBR+"stagedout"+c.getProgram ().getStandardOutput()));
		 */
		// }*/
		ArrayList<String> arguments = new ArrayList<String>();
		for (Arg arg : c.getProgram().getArgs()) {
			arguments.add(arg.getName());
			if (arg.getValues() != null) {
				for (String value : arg.getValues()) {
					if (value != null && !value.equals("")) {
						arguments.add(value);
					}
				}
			}
		}
		descriptionType.setArgument(arguments.toArray(new String[0]));
		// finally saves the RSL as a document in the current session
		XmlUtils.saveStringAsXMLDocument(
				RSLHelper.convertToString(descriptionType), SESSION_PATH
						+ session.getSessionID() + "/" + task.getName()
						+ "_rsl.xml");
		return descriptionType;
	}

	/**
	 * Generate the transfer request for staging purpose. Staging is required
	 * when we need to send result back to the invoking host
	 * 
	 * @param sourceURL
	 * @param destURL
	 * @return
	 */
	private TransferRequestType getStagingRequest(String sourceURL,
			String destURL) {
		// in bytes
		TransferRequestType request;
		try {
			TransferType[] transferArray = new TransferType[1];
			transferArray[0] = new TransferType();
			transferArray[0].setSourceUrl(sourceURL);
			transferArray[0].setDestinationUrl(destURL);
			RFTOptionsType rftOptions = new RFTOptionsType();
			rftOptions.setNotpt(new Boolean(false)); // yes = thirdPartyTransfer
			rftOptions.setBinary(new Boolean(true));
			rftOptions.setBlockSize(new Integer(16000)); // Block size in bytes
			rftOptions.setTcpBufferSize(new Integer(16000)); // TCP Buffer size
			request = new TransferRequestType();
			request.setRftOptions(rftOptions);
			request.setTransfer(transferArray);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
		return request;
		// request.setTransferCredentialEndpoint(eprDelegationResource);

	}

	/**
	 * Starts this computation task
	 */
	public synchronized void run() {
		try {
			state = State.CREATED;
			prepareForExecution();
			JobDescriptionType rsl = getRsl();
			String hostName = ((Computation) task).getProgram().getHostName();
			if (!hostName.contains(":")) {
				hostName += ":" + CONTAINER_PORT;
			}
			Utils.print(">>> Starting computational task \"" + task.getName()
					+ "\" on " + hostName);
			URL factoryUrl = ManagedJobFactoryClientHelper.getServiceURL(
					CONTAINER_PROTOCOL + hostName + "/" + GRAM_FACTORY)
					.getURL();
			String factoryType = ManagedJobFactoryConstants.FACTORY_TYPE.FORK;
			EndpointReferenceType factoryEndpoint = ManagedJobFactoryClientHelper
					.getFactoryEndpoint(factoryUrl, factoryType);
			submitRSL(factoryEndpoint, rsl);
			Utils.print("<<< Computational task \"" + task.getName() + "\" on "
					+ hostName + " ended");
		} catch (Exception e) {
			throw new RuntimeException("Error in execution", e);
		} finally {
			executionHandler.fireTaskResult(this);
		}
	}

	/**
	 * Delete and create folder and file
	 * 
	 * @throws IOException
	 */
	private void prepareForExecution() throws IOException {
		Computation c = (Computation) task;
		// File outfolder=new
		// File(c.getProgram().getOutputFolder()+"/"+session.getSessionID());
		File stderr = new File(KBR_PATH + c.getProgram().getStandardError());
		File stdout = new File(KBR_PATH + c.getProgram().getStandardOutput());
		// if(!outfolder.exists())
		// outfolder.mkdir();
		if (stderr.exists()) {
			stderr.delete();
		}
		if (stdout.exists()) {
			stdout.delete();
		}
	}

	@Override
	/**
	 * Error during execution. It is directly read from the standard out if Gramjob return no fault.
	 */
	public Throwable getErrorCause() throws Exception {
		if (job.getFault() != null) {
			if (job.getFault().getCause() == null) {
				int errorCode = job.getFault().getGt2ErrorCode();
				String errorMessage = "GRAM job GT2 error code = " + errorCode;
				return new LocalExecutionError(errorMessage);
			}
			return job.getFault().getCause();
		}
		Computation c = (Computation) task;
		String line, error = "";
		BufferedReader reader = new BufferedReader(new FileReader(KBR_PATH
				+ c.getProgram().getStandardError()));
		while ((line = reader.readLine()) != null) {
			error += line + "\n";
		}
		return new LocalExecutionError(error);
	}
}
