/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.execution;

import java.io.File;
import java.util.Iterator;

import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.Utils;
import kgrid.raems.service.impl.utils.ID;
import kgrid.raems.service.impl.utils.KDSClient;
import kgrid.raems.service.impl.utils.TransferInfo;
import kgrid.raems.service.impl.utils.TransferStatusListener;
import kgrid.raems.service.impl.utils.XmlUtils;
import kgrid.raems.service.impl.workflow.DataTransfer;
import kgrid.raems.service.impl.workflow.Resource;
import kgrid.raems.service.impl.workflow.Task;

/**
 * Execute a datatransfer using RFTService
 * 
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class DataTransferTask extends ExecutionThread implements
		TransferStatusListener, Consts {
	private int transferCount, completed;
	private RAEMSFileTransferManager rftJob;
	private String errorMessage = "Error unknown";

	public DataTransferTask(Task t, ExecutionHandler eh) {
		super(t, eh);
		transferCount = 0;
		completed = 0;
	}

	/**
	 * get a GSI URL from an host and a filesystem location
	 * 
	 * @param host
	 * @param physicalLocation
	 * @return
	 */
	private String getGsiUrl(String host, String physicalLocation) {
		if (host.contains(":")) {
			host = host.substring(0, host.indexOf(":"));
		}
		return GSIFTP_PROTOCOL + host + ":" + GSI_PORT + physicalLocation;
	}

	/**
	 * Start the transfer process
	 */
	public void run() {
		try {
			Utils.print(">>> Data transfer task \"" + task.getName()
					+ "\" started");
			state = State.CREATED;

			DataTransfer dataTransfer = (DataTransfer) task;

			Resource source = getDataTransferInput(dataTransfer);

			if (Utils.isLocalHost(source.getHostName())) {
				if (!new File(source.getPhysicalLocation()).exists()) {
					errorMessage = "Source file "
							+ source.getPhysicalLocation()
							+ " doesn't exist or has been prematurely deleted";
					Utils.print(errorMessage);
					state = State.FAILED;
					throw new RuntimeException(errorMessage);
				}
			} else {
				// TODO RAEMS Check whether the file exists on the remote host
			}

			Resource destination = getDataTransferOutput(dataTransfer);

			ID id = source.getId();
			String sourceResourceID = id.getStringID();

			/*
			 * Detect if this data transfer is needed. It's unnecessary when
			 * source and destination are the same. It could not be upgrade to
			 * different hosts cause we didn't want to prevent files overwriting
			 * (e.g. avoid transfer if a file with the same name as source
			 * resource exists on the destination host)
			 */
			if (source.equals(destination)) {
				// Transfer not necessary
				state = State.DONE;
				Utils.print("<<< Data transfer task \"" + task.getName()
						+ "\" not necessary");
				return;
			}

			// For each data transfer there's at least one transfer
			transferCount = 1;

			if (dataTransfer.isPublishResource()) {
				// if publishing is needed then this data transfer includes 2
				// transfers
				// TODO (-x-) RAEMS: don't publish if it's already published
				rftJob = new RAEMSFileTransferManager(
						session.getClientCredentials(), task.getName()
								+ "_metadata", this);
				transferCount++;
				String sourceUrlMetaData = getGsiUrl(source.getHostName(),
						KMR_PATH + source.getTitle());
				String destinationUrlMetaData = getGsiUrl(
						destination.getHostName(), DATATRANSFER_DOWNLOAD_PATH
								+ source.getTitle());

				TransferInfo[] tinfo = new TransferInfo[1];
				tinfo[0] = new TransferInfo(sourceUrlMetaData,
						destinationUrlMetaData);
				// **********************************************************************************************
				// System.out.println("RAEMS/IMPL/EXECUTION/DATATRANSFERTASK 1 "+task.getName()+": TRASFERIMENTO DI "+tinfo[0].getSourceURL());
				// String
				// notify="#STATE:"+task.getName()+":"+idResource+"$ACTIVATE DOWNLOAD - Source="+sourceUrlMetaData+" : Destination="+destinationUrlMetaData;
				// session.setResourceModel(notify);
				rftJob.performTransfer(sourceUrlMetaData,
						destinationUrlMetaData);
				// ReliableFileTransferManager takes care of both RFT resource
				// and Subscription resource destruction
				rftJob.destroy();
				if (rftJob.getFault() != null) {
					state = State.FAILED;
					throw new RuntimeException("Transfer from "
							+ tinfo[0].getSourceURL() + " to "
							+ tinfo[0].getDestinationURL() + " has failed",
							rftJob.getFault());
				} else {
					// Transfer from tinfo[0].getSourceURL() to
					// tinfo[0].getDestinationURL() succeeded

					// notify="#STATE:"+task.getName()+":"+idResource+"$DONE DOWNLOAD - Source="+sourceUrlMetaData+" : Destination="+destinationUrlMetaData;
					// session.setResourceModel(notify);

					// FIXME RAEMS: don't publish if the transfer has failed
					KDSClient client = new KDSClient(session);
					client.publish(dataTransfer.getDestination(),
							DATATRANSFER_DOWNLOAD_PATH + source.getTitle());
				}
			}
			/*
			 * check current status, that means check if a publishing operation
			 * had to be done and in this case if it was successful
			 */
			if (state.equals(ExecutionThread.State.FAILED)) {
				String notify = "#STATE:" + task.getName() + ":"
						+ sourceResourceID
						+ "$FAILED Data transfer failed for an unknown reason";
				session.setResourceModel(notify);
			} else {
				String sourceUrl = getGsiUrl(source.getHostName(),
						source.getPhysicalLocation());
				String destinationUrl = getGsiUrl(destination.getHostName(),
						destination.getPhysicalLocation());

				rftJob = new RAEMSFileTransferManager(
						session.getClientCredentials(), task.getName(), this);
				// System.out.println("RAEMS/IMPL/EXECUTION/DATATRANSFERTASK 2 "+task.getName()+": TRASFERIMENTO DI "+sourceUrl);
				String notify = "#STATE:" + task.getName() + ":"
						+ sourceResourceID + "$ACTIVE DataTransfer - Source="
						+ sourceUrl + " : Destination=" + destinationUrl;

				session.setResourceModel(notify);

				// throws UnexpectedReplyCodeException when doing MLST on
				// non-existing file/dir
				String description = rftJob.performTransfer(sourceUrl,
						destinationUrl);

				// Transfer Completed. Destroying rft resource
				// finally save the transfer description in the current session
				XmlUtils.saveStringAsXMLDocument(description, SESSION_PATH
						+ session.getSessionID() + "/" + task.getName()
						+ "_xfr.xml");
				// ReliableFileTransferManager takes care of both RFT resource
				// and Subscription resource destruction
				rftJob.destroy();
				notify = "#STATE:" + task.getName() + ":" + sourceResourceID
						+ "$DONE DataTransfer - Source=" + sourceUrl
						+ " : Destination=" + destinationUrl;
				session.setResourceModel(notify);
			}
			state = State.DESTROYED;
			Utils.print("<<< Data transfer task \"" + task.getName()
					+ "\" ended");
		} catch (Throwable t) {
			state = State.FAILED;
			Utils.print(t.getMessage());
			throw new RuntimeException(t);
		} finally {
			executionHandler.fireTaskResult(this);
		}

	}

	/**
	 * the input of this data transfer task is the first input of the associated
	 * data transfer
	 * 
	 * @param dataTransfer
	 * @return
	 */
	private Resource getDataTransferInput(DataTransfer dataTransfer) {
		Iterator<Resource> it = dataTransfer.getInputs().iterator();
		return it.next();
	}

	/**
	 * the output of this data transfer task is the first output of the
	 * associated data transfer
	 * 
	 * @param dataTransfer
	 * @return
	 */
	private Resource getDataTransferOutput(DataTransfer dataTransfer) {
		Iterator<Resource> it = dataTransfer.getOutputs().iterator();
		return it.next();
	}

	@Override
	/**
	 * Callback for TransferStatusListener
	 */
	public synchronized void statusChanged(RAEMSFileTransferManager transferJob) {
		try {
			// handle state
			RAEMSFileTransferManager.State jobStatus = transferJob.getStatus();
			if (jobStatus.equals(RAEMSFileTransferManager.State.ACTIVE)
					|| jobStatus.equals(RAEMSFileTransferManager.State.PENDING)) {
				state = State.ACTIVE;
			}
			if (jobStatus.equals(RAEMSFileTransferManager.State.FAILED)) {
				state = State.FAILED;
			}
			if (jobStatus.equals(RAEMSFileTransferManager.State.FINISHED)) {
				completed++;
				if (completed >= transferCount) {
					state = State.DONE;
				}
			}
		} catch (Throwable t) {
			state = State.FAILED;
			throw new RuntimeException("", t);
		} finally {
			executionHandler.fireTaskResult(this);
			notifyAll();
		}
	}

	@Override
	public synchronized Throwable getErrorCause() {
		if (rftJob == null) {
			return new RuntimeException(errorMessage);
		}
		return rftJob.getFault();
	}

}
