/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.execution;

import java.util.Collection;

import kgrid.raems.gui.NotificationHandler;
import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.Utils;
import kgrid.raems.service.impl.exception.RAEMSExecutionException;
import kgrid.raems.service.impl.execution.ExecutionThread.State;
import kgrid.raems.service.impl.utils.Session;
import kgrid.raems.service.impl.workflow.Computation;
import kgrid.raems.service.impl.workflow.DataTransfer;
import kgrid.raems.service.impl.workflow.Job;
import kgrid.raems.service.impl.workflow.Task;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class ExecutionHandler implements Consts {
	private WorkflowHandler wfHandler;
	private Job concreteExecPlan;
	private boolean proceed;
	private Session session;

	public ExecutionHandler(Job concreteEP, Session session) {
		wfHandler = new WorkflowHandler(concreteEP);
		this.session = session;
		concreteExecPlan = concreteEP;
	}

	public synchronized void execute() throws InterruptedException {
		proceed = true;
		while (!(wfHandler.successful() || wfHandler.error())) {
			Collection<String> execTasks = wfHandler.getExecutableTasks();
			Utils.print(execTasks.size() + " executable tasks");
			for (String taskName : execTasks) {
				Utils.print("\t+ " + taskName);
			}
			for (String taskName : execTasks) {
				Task t = concreteExecPlan.getTask(taskName);
				ExecutionThread execThread = null;
				if (t instanceof Computation) {
					execThread = new ComputationTask(t, this);
				} else if (t instanceof DataTransfer) {
					execThread = new DataTransferTask(t, this);
				}
				execThread.setSession(session);
				wfHandler.runTask(t);
				execThread.start();
			}
			proceed = false;
			while (!proceed) {
				wait();
			}
		}
		while (!wfHandler.completed()) {
			// Not all job are destroyed. Waiting for destruction
			wait();
		}
		Utils.print("Destruction Completed");
		if (wfHandler.successful()) {
			session.setResourceModel("Concrete Execution Plan completed Successfully");
		} else if (wfHandler.error()) {
			session.setResourceModel("Concrete Execution Plan couldn't complete because of some error(s) in its Tasks");
		}
	}

	public synchronized void fireTaskResult(ExecutionThread executionThread)
			throws RAEMSExecutionException {
		Task task = executionThread.getTask();
		State taskState = executionThread.getTaskState();
		try {
			notifyState(task, taskState);
			if (taskState.equals(ExecutionThread.State.DONE)) {
				proceed = true;
				wfHandler.completeTask(task);
			} else if (taskState.equals(ExecutionThread.State.FAILED)) {
				proceed = true;
				if (executionThread.getErrorCause() != null) {
					notifyError(task, "Caused by:"
							+ executionThread.getErrorCause().getMessage());
				} else {
					notifyError(task, "No error cause available");
				}
				wfHandler.failTask(task);
			} else if (taskState.equals(ExecutionThread.State.DESTROYED)) {
				wfHandler.destroyTask(task);
			}
		} catch (Throwable t) {
			proceed = true;
			throw new RAEMSExecutionException(t);
		} finally {
			if (proceed) {
				notifyAll();
			}
		}
	}

	private void notifyError(Task t, String msg) {
		session.setResourceModel(NotificationHandler.ERROR_NOTIF + t.getName()
				+ "$" + msg);
	}

	private void notifyState(Task t, State state) {
		session.setResourceModel(NotificationHandler.STATE_NOTIF + t.getName()
				+ "$" + state);
	}
}
