/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.execution;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;

import kgrid.raems.service.impl.workflow.Job;
import kgrid.raems.service.impl.workflow.Task;
import kgrid.raems.service.impl.workflow.TaskDAG;
import kgrid.raems.service.impl.workflow.constants.Constants;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class WorkflowHandler {
	private Job concreteEP;
	private HashSet<String> runningTasks, failedTasks, completedTasks,
			destroyedTasks;

	public WorkflowHandler(Job concreteEP) {
		runningTasks = new HashSet<String>();
		failedTasks = new HashSet<String>();
		destroyedTasks = new HashSet<String>();
		destroyedTasks.add(Constants.StartNode);
		destroyedTasks.add(Constants.EndNode);
		completedTasks = new HashSet<String>();
		completedTasks.add(Constants.StartNode);
		completedTasks.add(Constants.EndNode);
		this.concreteEP = concreteEP;
	}

	public Collection<String> getExecutableTasks() {
		TaskDAG dag = concreteEP.getDAG();
		LinkedList<String> execTasks = new LinkedList<String>();
		for (String currentTask : concreteEP.getTaskIDs()) {
			if (completedTasks.contains(currentTask)
					|| failedTasks.contains(currentTask)
					|| runningTasks.contains(currentTask)) {
				continue;
			}
			boolean addTask = true;
			for (String precedingTask : dag.precedingTasks(currentTask)) {
				if (!completedTasks.contains(precedingTask)) {
					addTask = false;
					break;
				}
			}
			if (addTask) {
				execTasks.add(currentTask);
			}
		}
		return execTasks;
	}

	public boolean error() {
		return !failedTasks.isEmpty();
	}

	public void failTask(Task t) {
		runningTasks.remove(t.getName());
		failedTasks.add(t.getName());
	}

	public boolean successful() {
		return getExecutableTasks().isEmpty() && failedTasks.isEmpty()
				&& runningTasks.isEmpty();
	}

	public void completeTask(Task t) {
		runningTasks.remove(t.getName());
		completedTasks.add(t.getName());
	}

	public void runTask(Task t) {
		runningTasks.add(t.getName());
	}

	public void destroyTask(Task t) {
		destroyedTasks.add(t.getName());
	}

	public boolean completed() {
		// if (errorTasks.size()>0)
		// return
		// return destroyedTasks.size()>=concreteEP.getAllTasks().size();
		return destroyedTasks.size() >= completedTasks.size()
				+ failedTasks.size()
				&& runningTasks.size() <= 0;
	}
}
