/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.scheduler;

import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.security.KeyException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.Utils;
import kgrid.raems.service.impl.utils.ID;
import kgrid.raems.service.impl.workflow.CloneableAbstractResource;
import kgrid.raems.service.impl.workflow.Computation;
import kgrid.raems.service.impl.workflow.DataTransfer;
import kgrid.raems.service.impl.workflow.Job;
import kgrid.raems.service.impl.workflow.Pair;
import kgrid.raems.service.impl.workflow.Program;
import kgrid.raems.service.impl.workflow.Resource;
import kgrid.raems.service.impl.workflow.ResourceCache;
import kgrid.raems.service.impl.workflow.Task;

import org.globus.wsrf.NoSuchResourceException;

/**
 * This class takes care of remote execution derived problems, including hidden
 * data transfer detection and explicit data transfer finalization
 * 
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class RemoteExecutionResolver implements Consts {
	private int clonedItems = 0;
	private int transferData = 0;
	private ResourceCache resourceCache;
	private Job abstractEP;

	private HashMap<String, Integer> transferCount;
	private HashMap<String, Integer> transferCountInput;

	/**
	 * set the job to be resolved
	 * 
	 * @throws NoSuchResourceException
	 */
	public void setAbstractExecutionPlan(Job abstractEP)
			throws NoSuchResourceException {
		this.abstractEP = abstractEP;
		this.resourceCache = abstractEP.getResourceCache();
		transferCount = new HashMap<String, Integer>();
		transferCountInput = new HashMap<String, Integer>();
	}

	/**
	 * Finalize an explicit data transfer. This includes the creation of an
	 * output resource
	 * 
	 * @param dataTransfer
	 * @throws KeyException
	 * @throws CloneNotSupportedException
	 * @throws UnknownHostException
	 */
	private void resolve(DataTransfer dataTransfer) throws KeyException,
			CloneNotSupportedException, UnknownHostException {
		// TODO (-x-) RAEMS: check publication...
		for (Resource input : dataTransfer.getInputs()) {
			// the real resource
			Resource source;
			if (input.isReferenced()) {
				source = resourceCache.getReferencedResource(input);
			} else {
				source = input;
			}
			// the source host for this data transfer is the host in which the
			// input resource is stored
			dataTransfer.setSource(source.getHostName());
			// the output resource is an exact copy of the input resource
			Resource cloned = createDataTransferOutput(source, "explicitDT",
					dataTransfer.getDestination());
			if (source.isPublished()) {
				dataTransfer.setPublishResource(true);
			}
			dataTransfer.addOutput(cloned);
			// TODO GESTIRE LE COMPUTAZIONI CHE REFERENZIANO L'INPUT DEL DT
			// if (!input.getId().isIDRef())
			// for (Resource r : resourceCache.getResourceReferences(input)) {
			// r.setId(new ID(cloned.getId().getStringID(), true));
			// }
		}
	}

	private List<Pair<DataTransfer, Computation>> resolve(Computation c)
			throws KeyException, CloneNotSupportedException,
			UnknownHostException {
		LinkedList<Pair<DataTransfer, Computation>> ret = new LinkedList<Pair<DataTransfer, Computation>>();
		Program program;
		if (c.getProgram().getId().isIDRef()) {
			program = (Program) resourceCache.getReferencedResource(c
					.getProgram());
		} else {
			program = c.getProgram();
		}

		List<Pair<Resource, Resource>> replace = new LinkedList<Pair<Resource, Resource>>();
		for (Resource i : c.getInputs()) {
			Resource input = null;
			if (i.getId().isIDRef()) {
				input = resourceCache.getReferencedResource(i);
			} else {
				input = i;
			}
			// create a data transfer node to the program host for every input
			// (if needed) .
			// if (isRemote(input, program)) {
			Pair<DataTransfer, Computation> dataTransferPair = createHiddenDataTransfer(
					c, program, input, "inputTransfer");
			Iterator<Resource> data_it = dataTransferPair.getFirst()
					.getOutputs().iterator();
			Resource dataTransferOutput = data_it.next();
			ret.add(dataTransferPair);
			Resource computationInput = dataTransferOutput;
			replace.add(new Pair<Resource, Resource>(i, computationInput));
			// }
		}
		for (Pair<Resource, Resource> pair : replace) {
			c.replaceResource(pair.getFirst(), pair.getSecond());
			String oldStringID = pair.getFirst().getId().getStringID(), newStringID = pair
					.getSecond().getId().getStringID();
			c.getProgram().getArgs().getArgumentByValue(oldStringID)
					.replaceValue(oldStringID, newStringID);
		}
		return ret;
	}

	// search for conflicts (program and data are on different hosts) and fix
	// 'em
	public void resolveJob() throws Exception {
		HashMap<Computation, List<Pair<DataTransfer, Computation>>> conflicts = new HashMap<Computation, List<Pair<DataTransfer, Computation>>>();
		HashMap<Computation, List<DataTransfer>> resultsTranfers = new HashMap<Computation, List<DataTransfer>>();
		// gather conflicts
		for (Task t : abstractEP.getAllTasks()) {
			if (t instanceof DataTransfer) {
				DataTransfer dataTransfer = (DataTransfer) t;
				resolve(dataTransfer);
			} else if (t instanceof Computation) {
				Computation c = (Computation) t;
				List<Pair<DataTransfer, Computation>> hiddenDT = resolve(c);
				resultsTranfers.put(c, handleResults(c));
				if (hiddenDT != null) {
					conflicts.put(c, hiddenDT);
				}
			}

		}
		// fix conflicts
		for (Computation c : conflicts.keySet()) {
			for (Pair<DataTransfer, Computation> pair : conflicts.get(c)) {
				// break an edge in the graph, insert a data transfer node in
				// between
				if (!pair.getSecond().equals(c)) {
					abstractEP.addBetween(pair.getSecond(), c, pair.getFirst());
				} else {
					// every edge entering in this node is broken and
					// interleaved with a data transfer node
					abstractEP.addBefore(pair.getFirst(), c);
				}
			}
		}
		// fix results
		for (Computation c : resultsTranfers.keySet()) {
			List<DataTransfer> transfers = resultsTranfers.get(c);
			for (DataTransfer dt : transfers) {
				abstractEP.addBetween(c, abstractEP.getTask("END"), dt);
			}
		}

	}

	/**
	 * clonedOutput.setName(name) transfer computation results to local KBR
	 * (final results) or data transfer location (partial results)
	 * 
	 * @param c
	 * @return
	 * @throws MalformedURLException
	 * @throws CloneNotSupportedException
	 * @throws UnknownHostException
	 */
	private List<DataTransfer> handleResults(Computation c)
			throws MalformedURLException, UnknownHostException,
			CloneNotSupportedException {
		List<DataTransfer> dts = new LinkedList<DataTransfer>();
		for (Resource r : c.getOutputs()) {

			// DataTransfer dataTransfer = new DataTransfer(c.getName()
			// + "_resultTransfer" + this.transferData++);

			Integer count = null;
			if (!transferCount.containsKey(c.getName())) {
				count = new Integer(0);
				transferCount.put(c.getName(), count);
			} else {
				count = transferCount.get(c.getName());
				int value = count.intValue();
				value++;
				count = new Integer(value);
				transferCount.put(c.getName(), count);
			}
			this.transferData++;
			DataTransfer dataTransfer = new DataTransfer(c.getName()
					+ "_resultTransfer" + count.intValue());

			dataTransfer.setSource(r.getHostName());
			URL destination = null;
			// if (abstractEP.isFinalTask(c)) {
			destination = new URL("file://" + KBR_PATH);
			// COMMENTED TO AVOID PARTIAL RESULT TRANSFERS
			// else
			// destination = new URL("file://"
			// + DATATRANSFER_DOWNLOAD_LOCATION);
			String destFileAddress = destination.getProtocol() + "://"
					+ destination.getHost() + destination.getFile();
			dataTransfer.setDestination(destFileAddress);
			dataTransfer.addInput(r);
			dataTransfer.addOutput(createDataTransferOutput(r,
					"result_transfer", destFileAddress));
			dts.add(dataTransfer);
			// }
		}
		return dts;
	}

	/**
	 * Creates an output for a data transfer. The destination path is determined
	 * using the available information: if the destination is specified as a
	 * complete url then the filename part of the URL is set to be the path,
	 * else if the input resource is published on the source host it will be
	 * published on the destination host and hence the path MUST be exactly the
	 * input physical location, if the input isn't published a default data
	 * transfer location is used Generated output will be a real resource and
	 * will therefore have a new ID
	 * 
	 * @throws UnknownHostException
	 */
	private Resource createDataTransferOutput(Resource input, String purpose,
			String destination) throws CloneNotSupportedException,
			UnknownHostException {
		// we start from an exact copy of the input
		Resource clonedOutput = ((CloneableAbstractResource) input).clone();

		clonedOutput.setId(new ID(clonedOutput.getId().getStringID() + "cloned"
				+ clonedItems++, false));
		String host = "", path = "";
		try {
			// if an url has been specified than the output is set to be in the
			// specified location
			URL url = new URL(destination);
			clonedOutput.setName(normalizedName(input.getName(), purpose,
					clonedItems++));
			path = url.getFile();
			if (!path.endsWith("/")) {
				path += "/";
			}
			path += clonedOutput.getName();
			host = url.getHost();
			if (host.equals("") || host == null) {
				host = Utils.HOST_NAME;
			}
		} catch (MalformedURLException e) {

			if (destination.contains(":")) {
				String before = destination;
				destination = destination
						.substring(0, destination.indexOf(":"));
				Utils.print("Destination changed from \"" + before + "\" to \""
						+ destination);
			}
			// else treat destination as an host and set transfer location to
			InetAddress hostInet = InetAddress.getByName(destination);
			host = hostInet.getHostName();
			if (input.isPublished()) { // the same as input location if it needs
				// to be published
				path = input.getPhysicalLocation();
			} else {// data transfer default location
				clonedOutput.setName(normalizedName(input.getName(), purpose,
						clonedItems++));
				path = DATATRANSFER_DOWNLOAD_PATH + clonedOutput.getName();
			}
		}
		clonedOutput.setConcrete(host, path);

		return clonedOutput;
	}

	private String normalizedName(String name, String purpose, int number) {
		int index = name.lastIndexOf(".");
		if (index > 0)
			return name.substring(0, index) + "-" + purpose + "_" + number
					+ name.substring(index, name.length());
		return name + "_" + number;

	}

	private Pair<DataTransfer, Computation> createHiddenDataTransfer(
			Computation c, Program program, Resource input, String purpose)
			throws CloneNotSupportedException, UnknownHostException {
		DataTransfer transferData = null;
		Computation owner = (Computation) resourceCache.getOwner(input);
		Integer count = null;
		if (!transferCountInput.containsKey(c.getName())) {
			count = new Integer(0);
			transferCountInput.put(c.getName(), count);
		} else {
			count = transferCountInput.get(c.getName());
			int value = count.intValue();
			value++;
			count = new Integer(value);
			transferCountInput.put(c.getName(), count);
		}
		this.transferData++;
		// transferData = new DataTransfer(new String(c.getName()
		// + "_"+purpose + this.transferData++));
		transferData = new DataTransfer(new String(c.getName() + "_" + purpose
				+ count.intValue()));
		// if (c.equals(owner)) {
		transferData.addInput(input);

		// }else{
		//
		// ReferencedData dataTransferInput = new ReferencedData();
		// dataTransferInput.setId(new ID(input.getId().getStringID(), true));
		// transferData.addInput(dataTransferInput);
		// }
		// System.out.println("RAEMS/IMPL/SCHEDULER/REMOTEEXECUTIONRESOLVER ===============>SOURCE="+input.getHostName()+" NAME="+input.getName()+" LOCATION="+input.getPhysicalLocation());
		transferData.setSource(input.getHostName());
		transferData.setDestination(program.getHostName());
		// System.out.println("RAEMS/IMPL/SCHEDULER/REMOTEEXECUTIONRESOLVER ===============>DESTINATION="+program.getHostName());

		Resource clonedOutput = createDataTransferOutput(input, purpose,
				program.getHostName());
		if (input.isPublished()) {
			transferData.setPublishResource(true);
		}
		transferData.addOutput(clonedOutput);
		return new Pair<DataTransfer, Computation>(transferData, owner);
	}

}
