/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.scheduler;

import java.rmi.RemoteException;
import java.security.KeyException;
import java.util.Iterator;
import java.util.Map;

import javax.xml.rpc.ServiceException;

import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.Utils;
import kgrid.raems.service.impl.exception.ConnectionException;
import kgrid.raems.service.impl.exception.DownloadException;
import kgrid.raems.service.impl.exception.GeneralException;
import kgrid.raems.service.impl.utils.KDSClient;
import kgrid.raems.service.impl.utils.KDSUrl;
import kgrid.raems.service.impl.utils.ResourceLookupInfo;
import kgrid.raems.service.impl.utils.XmlUtils;
import kgrid.raems.service.impl.workflow.Computation;
import kgrid.raems.service.impl.workflow.Job;
import kgrid.raems.service.impl.workflow.Program;
import kgrid.raems.service.impl.workflow.Resource;
import kgrid.raems.service.impl.workflow.Task;
import kgrid.raems.service.impl.workflow.plugin.Arg;

import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.encoding.SerializationException;
import org.globus.wsrf.query.QueryException;
import org.oasis.wsrf.lifetime.ResourceNotDestroyedFaultType;
import org.oasis.wsrf.lifetime.ResourceUnknownFaultType;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class Scheduler implements Consts {
	private KDSClient kdsClient;
	private Mapper mapper;

	public Scheduler(KDSClient client, Mapper mapper) {
		this.kdsClient = client;
		this.mapper = mapper;
	}

	public Job schedule(Job abstractEP) throws Exception {
		// TODO (-x-) RAEMS: first transfer the actual resource, then the
		// metadata
		queryAndUpdateResource(abstractEP);
		RemoteExecutionResolver resolver = new RemoteExecutionResolver();
		resolver.setAbstractExecutionPlan(abstractEP);
		resolver.resolveJob();
		for (Task t : abstractEP.getAllTasks()) {
			/* HashMap<Resource, Resource> toChange = */
			collectDataToRemove(t, abstractEP);
			/* removeReferencedData(t, toChange); */
			if (t instanceof Computation) {
				mapInput((Computation) t);
			}
		}
		return abstractEP;

	}

	private void mapInput(Computation c) {
		for (Resource input : c.getInputs()) {
			Arg inputArg = c.getProgram().getArgs()
					.getArgumentByValue(input.getId().getStringID());
			if (inputArg != null)
				inputArg.replaceValue(input.getId().getStringID(),
						input.getPhysicalLocation());
		}
		for (Resource output : c.getOutputs()) {
			Arg outputArg = c.getProgram().getArgs()
					.getArgumentByValue(output.getId().getStringID());
			if (outputArg != null)
				outputArg.replaceValue(output.getId().getStringID(),
						output.getPhysicalLocation());
		}

	}

	private void queryAndUpdateResource(Job abstractEP)
			throws ResourceNotDestroyedFaultType, ResourceUnknownFaultType,
			NoSuchResourceException, QueryException, SerializationException,
			RemoteException, ConnectionException, InterruptedException,
			ServiceException, GeneralException, DownloadException {

		Map<Resource, ResourceLookupInfo> resourceInfos = ResourceLookup
				.lookup(kdsClient, abstractEP);

		Map<Resource, KDSUrl> mappings = mapper.map(kdsClient, abstractEP,
				resourceInfos);

		for (Task t : abstractEP.getAllTasks()) {
			for (Resource r : t.getResources()) {
				KDSUrl url = mappings.get(r);
				if (url == null) {
					String name = "";
					if (!r.isReferenced()) {
						name = r.getName();
					}
					Utils.print("KDS URL for resource " + r.getId() + " "
							+ name + " not found");
				} else {
					String filepath = kdsClient.download(url);

					try {
						// if (!r.isPublished() && !r.isReferenced()) {
						if (r.isReferenced()) {
							// FIXME !!! If the resource is referenced the href
							// should be found looking for the resource whose id
							// is equal to the idref of the referenced resouce
						} else {
							/*
							 * Session session=kdsClient.getSession(); String
							 * nameResource; if(r.getName().equals(""))
							 * nameResource="Dataset"; else
							 * nameResource=r.getName(); String list="List$ ";
							 * ResourceLookupInfo rInfo=resourceInfos.get(r);
							 * String notify; if(rInfo!=null) { KDSResultSet
							 * resSet=rInfo.getResultSet(); for(int
							 * i=0;i<resSet.size();i++) {
							 * list+=((KDSUrl)resSet.get(i)).getURL();
							 * list+=" "; } notify="#"+nameResource+":"+list;
							 * session.setResourceModel(notify); }
							 * //notify="#Resource:"
							 * +nameResource+":$ "+url.getURL();
							 * //session.setResourceModel(notify);
							 */
							r.setPublicInformation(url.getURL(),
									url.getMetadataFileName());
							r.update(XmlUtils.loadFromFile(filepath
									+ url.getMetadataFileName()));
							String hostName = url.getHostName();
							if (url.getPort() != DEFAULT_PORT) {
								hostName += ":" + url.getPort();
								Utils.print(url.getURL()
										+ " not using the default port"
										+ hostName);
								Utils.print("New hostname = " + hostName);
							}
							r.setConcrete(hostName, r.getPhysicalLocation());
						}
					} catch (Exception e) {
						e.printStackTrace();
						throw new GeneralException(
								"Error updating resource(s) with concrete resource(s) infos");
					}
				}
			}
			if (t instanceof Computation) {
				Computation c = (Computation) t;
				for (Resource output : c.getOutputs()) {
					Program p = c.getProgram();
					output.setConcrete(p.getHostName(), p.getOutputFolder()
							+ output.getName());
				}
			}
		}
	}

	private void collectDataToRemove(Task t, Job abstractEP)
			throws CloneNotSupportedException, NoSuchResourceException,
			KeyException {
		// HashMap<Resource, Resource> toChange = new HashMap<Resource,
		// Resource>();
		Iterator<Resource> it = t.getResources().iterator();
		while (it.hasNext()) {
			Resource r = it.next();
			if (r.isReferenced()) {
				Resource found = abstractEP.getResourceCache()
						.getReferencedResource(r);
				if (found == null)
					Utils.print("No resource found with id: " + r.getId());
				// if (!(found instanceof CloneableAbstractResource))
				// throw new RuntimeException(
				// "A resource has been found but is not cloneable");
				// CloneableAbstractResource d = (CloneableAbstractResource)
				// found;
				// CloneableAbstractResource cloned =
				// (CloneableAbstractResource) d
				// .clone();
				// cloned.setId(new ID(cloned.getId().getStringID() + "cloned"
				// + (clonedItems++), false));
				t.replaceResource(r, found);
				// toChange.put(r, cloned);
			}
		}
		// return toChange;

	}
}
