/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.utils;

import java.rmi.RemoteException;

import javax.xml.namespace.QName;
import javax.xml.rpc.ServiceException;
import javax.xml.rpc.Stub;

import kgrid.kds.stubs.KDSFactory.CreateResource;
import kgrid.kds.stubs.KDSFactory.CreateResourceResponse;
import kgrid.kds.stubs.KDSFactory.KDSFactoryPortType;
import kgrid.kds.stubs.KDSFactory.service.KDSFactoryServiceAddressingLocator;
import kgrid.kds.stubs.KDSService.KDSPortType;
import kgrid.kds.stubs.KDSService.service.KDSServiceAddressingLocator;
import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.RAEMSListener;
import kgrid.raems.service.impl.Utils;
import kgrid.raems.service.impl.exception.ConnectionException;

import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.globus.axis.util.Util;
import org.globus.gsi.GSIConstants;
import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.WSNConstants;
import org.globus.wsrf.encoding.SerializationException;
import org.globus.wsrf.impl.notification.SubscriptionCreationException;
import org.globus.wsrf.impl.security.authorization.Authorization;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.descriptor.ClientSecurityDescriptor;
import org.oasis.wsn.Subscribe;
import org.oasis.wsn.SubscribeResponse;
import org.oasis.wsn.SubscriptionManager;
import org.oasis.wsn.TopicExpressionType;
import org.oasis.wsn.WSBaseNotificationServiceAddressingLocator;
import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.lifetime.ResourceNotDestroyedFaultType;
import org.oasis.wsrf.lifetime.ResourceUnknownFaultType;

/**
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class KDSConnectionUtils implements Consts {
	private static Object authVal = GSIConstants.SIGNATURE;
	private static Authorization authzVal = HostAuthorization.getInstance();

	/**
	 * Initialize KDS Resource and get the KDS Instance Service Port Type.
	 */
	public static KDSPortType getKDSPortType() throws ConnectionException {
		try {
			KDSFactoryServiceAddressingLocator factoryLocator = new KDSFactoryServiceAddressingLocator();

			KDSServiceAddressingLocator instanceLocator = new KDSServiceAddressingLocator();

			String kdsFactoryURI = CONTAINER_PROTOCOL + Utils.HOST_IP + ":"
					+ CONTAINER_PORT + "/" + KDS_FACTORY;

			EndpointReferenceType factoryEPR, instanceEPR;
			KDSFactoryPortType kdsFactory;
			KDSPortType kds;

			// Get factory portType
			factoryEPR = new EndpointReferenceType();
			factoryEPR.setAddress(new Address(kdsFactoryURI));
			kdsFactory = factoryLocator.getKDSFactoryPortTypePort(factoryEPR);

			// Create resource and get endpoint reference of
			// KDS-Resource.
			// This resource is our "instance".
			CreateResourceResponse createResponse = kdsFactory
					.createResource(new CreateResource());
			instanceEPR = createResponse.getEndpointReference();

			// Get instance PortType
			kds = instanceLocator.getKDSPortTypePort(instanceEPR);

			return kds;
		} catch (Exception e) {
			throw new ConnectionException("Error connecting to local KDS.", e);
		}
	}

	public static KDSPortType getKDSPortType(String hostIP)
			throws ConnectionException {
		try {
			KDSFactoryServiceAddressingLocator factoryLocator = new KDSFactoryServiceAddressingLocator();

			KDSServiceAddressingLocator instanceLocator = new KDSServiceAddressingLocator();

			String KDS_FactoryURI = CONTAINER_PROTOCOL + hostIP + ":"
					+ CONTAINER_PORT + "/" + KDS_FACTORY;

			EndpointReferenceType factoryEPR, instanceEPR;
			KDSFactoryPortType kdsFactory;
			KDSPortType kds;

			// Get factory portType
			factoryEPR = new EndpointReferenceType();
			factoryEPR.setAddress(new Address(KDS_FactoryURI));
			kdsFactory = factoryLocator.getKDSFactoryPortTypePort(factoryEPR);

			// Create resource and get endpoint reference of
			// KDS-Resource.
			// This resource is our "instance".
			CreateResourceResponse createResponse = kdsFactory
					.createResource(new CreateResource());
			instanceEPR = createResponse.getEndpointReference();

			// Get instance PortType
			kds = instanceLocator.getKDSPortTypePort(instanceEPR);

			return kds;
		} catch (Exception e) {
			throw new ConnectionException("Error connecting to KDS on host:"
					+ hostIP + ".", e);
		}
	}

	/**
	 * Subscribe the specified listener as consumer of notification 'KDSModel'
	 * topic of the specified KDS Service
	 * 
	 * @param listener
	 *            consumer of the notification
	 * @param kds
	 *            KDSPortType of the KDS Service whose the topic is a resource
	 *            property
	 * @throws SubscriptionCreationException
	 */
	public static SubscribeResponse subscribeToKDSModelTopic(
			RAEMSListener listener, KDSPortType kds)
			throws SubscriptionCreationException {
		try {

			// Subscribing the RAEMSListener to the KDS_Model topic

			// 1: set up the notification consumer
			// (the listener will be set up as a notification consumer)
			NotificationConsumerManager consumer = null;

			QName KDS_RP_MODEL = new QName(KDS_NAMESPACE, "KDSModel");

			// 2: set the 'listener' as a 'notification consumer'
			consumer = NotificationConsumerManager.getInstance();
			consumer.startListening();
			EndpointReferenceType consumerEPR = consumer
					.createNotificationConsumer(listener);

			// 3: create (and opportunetely set) the remote subscribe request
			// Create the request to the remote Subscribe() call
			Subscribe request = new Subscribe();
			// Must the notification be delivered using the Notify
			// operation?
			request.setUseNotify(Boolean.TRUE);
			// Indicate what the consumer client's EPR is
			request.setConsumerReference(consumerEPR);

			// 4: create the topic to be subscribed to
			// The TopicExpression specifies what topic we want to
			// subscribe to
			TopicExpressionType topicExpression = new TopicExpressionType();
			topicExpression.setDialect(WSNConstants.SIMPLE_TOPIC_DIALECT);
			topicExpression.setValue(KDS_RP_MODEL);
			request.setTopicExpression(topicExpression);

			// 6: send the subscribe request to the notification producer (KDS)
			// IMPORTANT: the consumer DASListener is subscribed to the
			// 'KDS_Model'
			// topic
			SubscribeResponse response = kds.subscribe(request);
			return response;
		} catch (Exception e) {
			throw new SubscriptionCreationException("Error subscribing to KDS",
					e);
		}
	}

	public static void destroySubscription(SubscribeResponse response)
			throws ServiceException, SerializationException,
			ResourceNotDestroyedFaultType, ResourceUnknownFaultType,
			RemoteException {
		EndpointReferenceType notificationEndpoint = response
				.getSubscriptionReference();
		if (notificationEndpoint != null) {
			SubscriptionManager subscriptionPort = new WSBaseNotificationServiceAddressingLocator()
					.getSubscriptionManagerPort(notificationEndpoint);
			setSecurity((Stub) subscriptionPort);
			subscriptionPort.destroy(new Destroy());
		}
	}

	private static void setSecurity(Stub stub) {
		stub._setProperty("clientDescriptor", getClientSecDesc());
	}

	private static ClientSecurityDescriptor getClientSecDesc() {
		ClientSecurityDescriptor desc = new ClientSecurityDescriptor();
		desc.setGSITransport((Integer) authVal);
		Util.registerTransport();
		desc.setAuthz(authzVal);
		return desc;
	}

}
