/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.utils;

import java.rmi.RemoteException;

import javax.xml.rpc.ServiceException;

import kgrid.kds.stubs.KDSService.ArrayOfString;
import kgrid.kds.stubs.KDSService.KDSPortType;
import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.RAEMSListener;
import kgrid.raems.service.impl.exception.ConnectionException;

import org.globus.wsrf.encoding.SerializationException;
import org.oasis.wsn.SubscribeResponse;
import org.oasis.wsrf.lifetime.Destroy;
import org.oasis.wsrf.lifetime.ResourceNotDestroyedFaultType;
import org.oasis.wsrf.lifetime.ResourceUnknownFaultType;

/**
 * This is a client for a local KDS service thus meaning the KDS must be located
 * on the same container where the RAEMS service runs. In particular, this
 * client just queries the KDS for the information needed to create a Concrete
 * Execution Plan
 * 
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class KDSQueryClient implements Consts {

	private Session session = null;

	public KDSQueryClient(Session session) {
		this.session = session;
	}

	public Session getSession() {
		return session;
	}

	/**
	 * makes all the queries needed to complete the Abstract Execution Plan and
	 * puts the results in a KDSResultSet Object
	 * 
	 * @return the results of all the queries done so far
	 * @throws ConnectionException
	 * @throws InterruptedException
	 * @throws ServiceException
	 * @throws RemoteException
	 * @throws SerializationException
	 * @throws ResourceUnknownFaultType
	 * @throws ResourceNotDestroyedFaultType
	 * 
	 */
	public KDSResultSet makeQuery(Query q) throws ConnectionException,
			InterruptedException, ResourceNotDestroyedFaultType,
			ResourceUnknownFaultType, SerializationException, RemoteException,
			ServiceException {
		RAEMSListener raemsListener = new RAEMSListener(this);
		KDSPortType kds;
		kds = KDSConnectionUtils.getKDSPortType();
		SubscribeResponse response = KDSConnectionUtils
				.subscribeToKDSModelTopic(raemsListener, kds);
		ArrayOfString result = null;
		// result = kds.searchResource(q.getArrayOfString());
		SearchThread searchThread = new SearchThread(kds, q);
		searchThread.start();
		synchronized (this) {
			this.wait();
		}
		KDSConnectionUtils.destroySubscription(response);
		kds.destroy(new Destroy());
		if (searchThread.getFault() != null) {
			throw searchThread.getFault();
		}
		result = searchThread.getResult();
		if (result.getX() == null) {
			return null;
		}
		return new KDSResultSet(result);
	}
}

class SearchThread extends Thread implements KDSClientThread, Consts {
	private ArrayOfString result;
	private KDSPortType kds;
	private Query query;
	private RemoteException fault = null;

	public SearchThread(KDSPortType kdsPort, Query q) {
		this.kds = kdsPort;
		this.query = q;
		result = null;
	}

	public void run() throws RuntimeException {
		try {
			result = kds.searchResource(query.getArrayOfString());
			if (result.getX() != null && result.getX()[0].equals("ERROR")) {
				/*
				 * The workflow execution must be terminated because a specific
				 * resource was not found
				 */
				fault = new RemoteException(result.getX()[1]);
			}
		} catch (RemoteException e) {
			fault = e;
		}
	}

	public ArrayOfString getResult() {
		return result;
	}

	public RemoteException getFault() {
		return fault;
	}
}
