/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.utils;

import java.io.IOException;

import kgrid.raems.service.impl.Consts;

import org.globus.wsrf.container.ServiceHost;

/**
 * 
 * Represents a KDS URL. The address must be composed in this way:
 * "kds://hostname/filename.extension"
 * 
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class KDSUrl implements Consts {

	private String host;
	private int port;
	private String metadata;
	private String name;

	/**
	 * 
	 * @param address
	 * @throws IllegalArgumentException
	 *             if address does not comply with the KDS URL structure
	 * @throws IOException
	 */
	public KDSUrl(String address) throws IllegalArgumentException {
		if (address == null || !address.startsWith(KDS_PROTOCOL)) {
			throw new IllegalArgumentException();
		}
		int k = address.indexOf("/", KDS_PROTOCOL.length());
		if (k == -1) {
			throw new IllegalArgumentException();
		}
		host = address.substring(KDS_PROTOCOL.length(), k);
		if (host.contains(":")) {
			port = Integer.valueOf(host.substring(host.indexOf(":") + 1));
			host = host.substring(0, host.indexOf(":"));
		} else {
			port = DEFAULT_PORT;
		}
		metadata = address.substring(k + 1);
		k = metadata.lastIndexOf(".");
		if (k == -1) {
			throw new IllegalArgumentException();
		}
		name = metadata.substring(0, k);
	}

	public int getPort() {
		return port;
	}

	public String getName() {
		return name;
	}

	public String getHostName() {
		return host;
	}

	public String getMetadataFileName() {
		return metadata;
	}

	public void setMetadataFileName(String metadata) {
		this.metadata = metadata;
	}

	public String getURL() {
		String portString = "";
		if (port != DEFAULT_PORT) {
			portString = ":" + port;
		}
		return KDS_PROTOCOL + host + portString + "/" + metadata;
	}

	@Override
	public boolean equals(Object o) {
		if (!(o instanceof KDSUrl)) {
			return false;
		}
		KDSUrl url = (KDSUrl) o;
		return url.host.equals(host) && url.metadata.equals(metadata)
				&& port == url.port;
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 31 * hash + host.hashCode();
		hash = 31 * hash + metadata.hashCode();
		hash = 31 * hash + port;
		return hash;
	}

}
