/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.workflow;

import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import kgrid.raems.service.impl.workflow.constants.Constants;
import kgrid.raems.service.impl.workflow.constants.JobConstants;

import org.w3c.dom.Element;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class Computation extends Task {

	private static final long serialVersionUID = 1L;
	private List<Resource> inputs;
	private Program program;
	private List<Resource> outputs;

	public Computation(String ID) {
		super(ID);
		inputs = new LinkedList<Resource>();
		program = null;
		outputs = new LinkedList<Resource>();
	}

	public List<Resource> getInputs() {
		return inputs;
	}

	public void addInput(Resource input) {
		if (input.getId() == null)
			throw new RuntimeException(input.getClass() + " must have an id");
		this.inputs.add(input);
	}

	public Program getProgram() {
		return program;
	}

	public void setProgram(Program program) {
		if (program.getId() == null)
			throw new RuntimeException("program must have an id");
		this.program = program;
	}

	public boolean isAbstract() {
		boolean input = false;
		for (Resource d : inputs)
			if (d.isAbstract())
				input = true;

		return input || program.isAbstract();
	}

	public String toString() {
		String s = "{\n\tComputation=" + getName() + "\n\tInput: "
				+ Arrays.toString(inputs.toArray()) + "\n\tProgram: "
				+ program.toString();
		if (outputs != null && outputs.size() > 0) {
			s += "\n\tOutput: " + Arrays.toString(outputs.toArray()) + "\n}";
		}
		return s;
	}

	public void addOutput(Resource output) {
		if (output.getId() == null)
			throw new RuntimeException("output must have an id");
		this.outputs.add(output);
	}

	public List<Resource> getOutputs() {
		return outputs;
	}

	public void replaceResource(Resource from, Resource to) {
		for (Resource input : inputs) {
			if (input.getId().equals(from.getId())) {
				inputs.remove(input);
				addInput(to);
				return;
			}
		}

		for (Resource output : outputs) {
			if (output.getId().equals(from.getId())) {
				outputs.remove(output);
				addOutput(to);
				return;
			}
		}
		if (program.getId().equals(from.getId())) {
			program = (Program) to;
		}

	}

	public Element serialize() throws ParserConfigurationException {
		Serializer rootSer = super.serializer();
		Serializer serializer = new Serializer(JobConstants.Computation);
		// input serialization
		for (Resource input : inputs) {
			Serializer input_ser = new Serializer(JobConstants.Input);

			input_ser.addAsAttribute(Constants.Href, input.getHref());
			input_ser.addAsAttribute(Constants.Title, input.getTitle());
			input_ser.addAsChildElement(input.serialize());
			serializer.addAsChildElement(input_ser.serializedElement());
		}
		// program serialization
		serializer.addAsChildElement(program.serialize());
		// output serialization
		for (Resource output : outputs) {
			if (output != null) {
				Serializer output_ser = new Serializer(JobConstants.Output);
				output_ser.addAsAttribute(Constants.Href, output.getHref());
				output_ser.addAsAttribute(Constants.Title, output.getTitle());
				output_ser.addAsChildElement(output.serialize());
				serializer.addAsChildElement(output_ser.serializedElement());
			}
		}
		rootSer.addAsChildElement(serializer.serializedElement());
		return rootSer.serializedElement();
	}

	public Collection<Resource> getResources() {
		LinkedList<Resource> ret = new LinkedList<Resource>();
		ret.addAll(inputs);
		ret.addAll(outputs);
		ret.add(program);
		return ret;
	}
}
