/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.workflow;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import kgrid.raems.service.impl.exception.NoSuchTaskException;
import kgrid.raems.service.impl.workflow.constants.Constants;
import kgrid.raems.service.impl.workflow.constants.JobConstants;

import org.globus.wsrf.NoSuchResourceException;
import org.w3c.dom.Document;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public final class Job implements Serializable {
    static final long             serialVersionUID = 5302651081024329415L;
    private boolean               dirty            = false;
    private TaskDAG               dag;
    // private HashMap<ID, Resource> resources;
    private ResourceCache         resources;
    private HashMap<String, Task> tasks;

    public Job(int DAG_TYPE) {
        resources = new ResourceCache();
        tasks = new HashMap<String, Task>();
        dag = SchedulerFactory.createTaskDAG(DAG_TYPE);
    }

    public void addAfter(Task child, Task parent) {
        addTask(child);
        Collection<String> subsequentTasks = dag.subsequentTasks(parent
                .getName());
        for (String id : subsequentTasks) {
            dag.removePrecedence(parent.getName(), id);
            link(child.getName(), id);
        }
        link(child.getName(), parent.getName());
    }

    public void addAfter(List<? extends Task> toAdd, Task parent) {
        addAllTasks(toAdd);
        Collection<String> subs = dag.subsequentTasks(parent.getName());
        for (String id : subs) {
            dag.removePrecedence(parent.getName(), id);
            for (Task t : toAdd) {
                link(parent.getName(), t.getName());
                link(t.getName(), id);
            }
        }
    }

    public void addBefore(Task toAdd, Task toPrecede) {
        addTask(toAdd);
        Collection<String> precs = dag.precedingTasks(toPrecede.getName());
        for (String id : precs) {
            dag.removePrecedence(id, toPrecede.getName());
            link(id, toAdd.getName());
        }

        link(toAdd.getName(), toPrecede.getName());

    }

    public void addBefore(List<? extends Task> toAdd, Task toPrecede) {
        addAllTasks(toAdd);
        Collection<String> precs = dag.precedingTasks(toPrecede.getName());
        for (String id : precs) {
            dag.removePrecedence(id, toPrecede.getName());
            for (Task t : toAdd) {
                link(id, t.getName());
                link(t.getName(), toPrecede.getName());
            }
        }

    }

    public void addTask(Task t) {
        tasks.put(t.getName(), t);
        dirty = true;
    }

    public void addBetween(Task t1, Task t2, Task toAdd) {
        addTask(toAdd);
        dag.removePrecedence(t1.getName(), t2.getName());
        link(t1.getName(), toAdd.getName());
        link(toAdd.getName(), t2.getName());

    }

    public void addAllTasks(List<? extends Task> toAdd) {
        for (Task t : toAdd) {
            tasks.put(t.getName(), t);
        }
        dirty = true;
    }

    public Collection<Task> getAllTasks() {
        return tasks.values();
    }

    public TaskDAG getDAG() {
        return dag;
    }

    public ResourceCache getResourceCache() throws NoSuchResourceException {
        if (dirty) {
            resources.buildResourceCache(this);
        }
        dirty = false;
        return resources;
    }

    public Task getTask(String taskID) {
        return tasks.get(taskID);
    }

    public Collection<String> getTaskIDs() {
        return new ArrayList<String>(tasks.keySet());
    }

    public boolean isAbstract() {

        for (Task t : tasks.values()) {
            if (t.isAbstract())
                return true;
        }
        return false;
    }

    public void link(String from, String to) {
        if (!tasks.containsKey(from) || !tasks.containsKey(to))
            throw new NoSuchTaskException("task does not exists");
        dag.addPrecedence(from, to);
    }

    // public void removeTask(String taskID) {
    // Task t = tasks.remove(taskID);
    // for (Resource r : t.getResources()) {
    // resources.remove(r.getId());
    // }
    // dag.removeTask(taskID);
    // }

    public int size() {
        return tasks.size();
    }

    public String toString() {
        String s = "Job\n";
        Collection<Task> listTask = tasks.values();
        Iterator<Task> it = listTask.iterator();
        while (it.hasNext()) {
            Task task = it.next();
            s += task.toString() + "\n";

        }
        return s + ", DAG:" + dag;
    }

    public Document serializeConcreteExecutionPlan()
            throws ParserConfigurationException {
        Serializer serializer = new Serializer("ConcreteExecutionPlan", "ep",
                "http://grid.deis.unical.it/kgrid/EPMSService");
        for (Task t : tasks.values())
            serializer.addAsChildElement(t.serialize());
        for (String task : dag.tasks()) {
            for (String subseq : dag.subsequentTasks(task)) {
                Serializer link_ser = new Serializer(JobConstants.TaskLink);
                link_ser.addAsAttribute(Constants.EpFrom, task);
                link_ser.addAsAttribute(Constants.EpTo, subseq);
                serializer.addAsChildElement(link_ser.serializedElement());
            }
        }
        return serializer.serializedDocument();
    }

    public boolean isFinalTask(Task t) {
        return dag.precedes(t.getName(), "END");
    }

}
