/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.workflow;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class TaskDAGMatrixImpl implements TaskDAG {
    static final long                serialVersionUID = 928700636651980252L;

    private HashMap<String, Integer> taskIDs_indexes;
    private HashMap<Integer, String> indexes_taskIDs;
    private boolean[][]              precedences;
    private int                      lastTaskIndexUsed;

    public TaskDAGMatrixImpl() {
        taskIDs_indexes = new HashMap<String, Integer>();
        indexes_taskIDs = new HashMap<Integer, String>();
        precedences = new boolean[100][100];
        for (int i = 0; i < 100; i++)
            for (int j = 0; j < 100; j++)
                precedences[i][j] = false;
        lastTaskIndexUsed = -1;
    }

    private int taskIndex(String taskID) {
        Integer o = taskIDs_indexes.get(taskID);
        return (o == null ? -1 : o);
    }

    private String taskID(int taskIndex) {
        return indexes_taskIDs.get(taskIndex);
    }

    public boolean precedes(String taskID1, String taskID2) // automatically
    // adds taskID1 and
    // taskID2 if not
    // existing
    {
        addTask(taskID1);
        addTask(taskID2);
        return precedences[taskIndex(taskID1)][taskIndex(taskID2)];
    }

    public void addPrecedence(String taskID1, String taskID2) // automatically
    // adds taskID1
    // and taskID2
    // if not
    // existing
    {
        addTask(taskID1);
        addTask(taskID2);
        precedences[taskIndex(taskID1)][taskIndex(taskID2)] = true;
    }

    private void doubleSize() {
        int currentSize = precedences[0].length;
        int newSize = currentSize * 2;
        boolean[][] newMatrix = new boolean[newSize][newSize];
        for (int i = 0; i < newSize; i++)
            for (int j = 0; j < newSize; j++) {
                if (i < currentSize && j < currentSize)
                    newMatrix[i][j] = precedences[i][j];
                else
                    newMatrix[i][j] = false;
            }
        precedences = newMatrix;
    }

    public void removePrecedence(String taskID1, String taskID2) // automatically
    // adds
    // taskID1
    // and
    // taskID2
    // if not
    // existing
    {
        addTask(taskID1);
        addTask(taskID2);
        precedences[taskIndex(taskID1)][taskIndex(taskID2)] = false;
    }

    public Collection<String> subsequentTasks(String taskID) {
        ArrayList<String> ret = new ArrayList<String>();
        int taskIndex = taskIndex(taskID);
        if (taskIndex != -1) {
            for (int j = 0; j < lastTaskIndexUsed; j++)
                if (precedences[taskIndex][j])
                    ret.add(taskID(j));
            return ret;
        }
        return new ArrayList<String>();
    }

    public Collection<String> precedingTasks(String taskID) {
        ArrayList<String> ret = new ArrayList<String>();
        int taskIndex = taskIndex(taskID);
        if (taskIndex != -1) {
            for (int i = 0; i < lastTaskIndexUsed; i++)
                if (precedences[i][taskIndex])
                    ret.add(taskID(i));
            return ret;
        }
        return new ArrayList<String>();
    }

    public void addTask(String taskID) {
        if (taskIDs_indexes.get(taskID) == null) {
            int taskIndex = ++lastTaskIndexUsed;
            taskIDs_indexes.put(taskID, taskIndex);
            indexes_taskIDs.put(taskIndex, taskID);
            if (taskIndex == precedences[0].length)
                doubleSize();
        }
    }

    public void removeTask(String taskID) {
        int ind = taskIndex(taskID);
        if (ind != -1) {
            indexes_taskIDs.remove(ind);
            taskIDs_indexes.remove(taskID);
            for (int p = 0; p < lastTaskIndexUsed; p++) {
                precedences[ind][p] = false;
                precedences[p][ind] = false;
            }
        }
    }

    public Collection<String> tasks() {
        return taskIDs_indexes.keySet();
    }

    public String toString() {
        String ret = "[DAG [Tasks:";
        for (String s : taskIDs_indexes.keySet())
            ret += s + " ";
        ret += "] [Precedences:";
        for (int i = 0; i < lastTaskIndexUsed; i++)
            for (int j = 0; j < lastTaskIndexUsed; j++)
                if (precedences[i][j])
                    ret += taskID(i) + "->" + taskID(j) + " ";
        ret += "] ]";
        ;
        return ret;
    }
}
