/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.workflow.plugin;

import java.util.HashSet;
import java.util.Set;

import javax.xml.parsers.ParserConfigurationException;

import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.utils.Query;
import kgrid.raems.service.impl.workflow.ArgumentList;
import kgrid.raems.service.impl.workflow.Deserializer;
import kgrid.raems.service.impl.workflow.PluginDescription;
import kgrid.raems.service.impl.workflow.Program;
import kgrid.raems.service.impl.workflow.Property;
import kgrid.raems.service.impl.workflow.Serializer;
import kgrid.raems.service.impl.workflow.constants.Constants;
import kgrid.raems.service.impl.workflow.constants.DataMiningSoftwareConstants;

import org.globus.wsrf.query.QueryException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class DataMiningSoftware extends Program implements Consts {

	private Set<UniqueProperty> description;
	private Set<UniqueProperty> usage; // contains the
										// type of file,
										// e.g.
	// executable, the path etc.
	private ArgumentList args; // contains the
								// list of argument
	// that
	// can be used
	private DataMiningSoftwareQueryConstructor query_constr;

	public DataMiningSoftware() {
		super();
		args = new ArgumentList();
		description = new HashSet<UniqueProperty>();
		usage = new HashSet<UniqueProperty>();
		query_constr = new DataMiningSoftwareQueryConstructor(this);
	}

	public void addDescriptionProperty(UniqueProperty property) {
		description.add(property);
	}

	public void addUsageProperty(UniqueProperty property) {
		usage.add(property);
	}

	public Set<UniqueProperty> getDescription() {
		return description;
	}

	public void setDescription(Set<UniqueProperty> description) {
		this.description = description;
	}

	public Set<UniqueProperty> getUsage() {
		return usage;
	}

	public void setUsage(Set<UniqueProperty> usage) {
		this.usage = usage;
	}

	public void addArgsProperty(Arg arg) {
		args.add(arg);
	}

	public ArgumentList getArgs() {
		return args;
	}

	public void update(Document document) throws Exception {
		PluginDescription pluginDescription = new PluginDescription(
				"DataMiningSoftware");
		Deserializer deserializer = new Deserializer(pluginDescription);

		DataMiningSoftware r = (DataMiningSoftware) deserializer
				.deserialize(document);
		this.description = r.description;
		this.usage = r.usage;
		this.name = r.name;
		int mandatoryCount = 0;
		ArgumentList myargs = null;
		for (Arg arg : r.getArgs()) {
			myargs = this.args.getArgumentByName(arg.getName());
			if (myargs.size() <= 0 && arg.isMandatory()) {
				Arg newMyArg = new Arg();
				newMyArg.setName(arg.getValues().get(0));
				newMyArg.addValue(arg.getDefault());
				this.args.add(mandatoryCount++, newMyArg);
			} else if (myargs.size() > 0) {
				if (arg.isMultivalue()) {
					Arg newMyArg = new Arg();
					newMyArg.setMultivalue(true);
					newMyArg.setName(arg.getValues().get(0));
					for (Arg a : myargs) {
						if (!a.getValues().get(0).equals("")
								&& !arg.getType().equals("none"))
							newMyArg.addValue(a.getValues().get(0));
						else
							newMyArg.addValue(arg.getDefault());
						this.args.remove(a);
					}
					this.args.add(newMyArg);
				} else {
					for (Arg a : myargs) {
						a.setName(arg.getValues().get(0));
					}
				}
			}
		}
	}

	public String toString() {
		String s = "[DataMiningSoftware\n";
		if (href != null && !href.equals(""))
			s += "(href: " + href + ")\n";
		if (id != null)
			s += id + "\n";
		if (title != null && !title.equals(""))
			s += "(title: " + title + ")\n";
		if (name != null && !name.equals(""))
			s += "(name: " + name + ")\n";
		if (description != null && description.size() > 0) {
			s += "(Description:\n";
			for (Property p : description)
				s += p.toString() + "\n";
		}
		s += ")\n";
		if (usage != null && usage.size() > 0) {
			s += "(usage:";
			for (Property p : usage)
				s += p.toString() + "\n";
		}
		s += ")\n";
		if (args != null && args.size() > 0) {
			s += "(args:\n";
			for (Arg arg : args)
				s += arg.toString() + "\n";
		}
		return s += ")]";

	}

	public Query getQuery() throws QueryException {
		try {
			return query_constr.getQuery();
		} catch (Exception e) {
			throw new QueryException("Error constructing query for resource "
					+ this.getId() + ".", e);
		}
	}

	public String getPhysicalLocation() {
		for (Property p : this.usage) {
			if (p.getName().equals(Constants.LOCATION))
				return p.getValue();
		}
		return "";
	}

	public boolean isReferenced() {
		return false;
	}

	public String getHostName() {
		for (Property p : usage) {
			if (p.getName().equals(Constants.HOSTNAME)) {
				return p.getValue();
			}
		}
		return "";
	}

	public void setConcrete(String hostname, String path) {
		UniqueProperty host = new UniqueProperty(), loc = new UniqueProperty();
		host.setName(Constants.HOSTNAME);
		host.setValue(hostname);
		loc.setName(Constants.LOCATION);
		loc.setValue(path);
		if (!usage.contains(host)) {
			addUsageProperty(host);
		} else {
			for (Property p : this.usage) {
				if (p.getName().equals(Constants.HOSTNAME))
					p.setValue(hostname);
			}
		}
		if (!usage.contains(loc)) {
			addUsageProperty(loc);
		} else {
			for (Property p : this.usage) {
				if (p.getName().equals(Constants.LOCATION))
					p.setValue(path);
			}
		}
	}

	public String getOutputFolder() {
		for (Property p : usage) {
			if (p.getName().equals(Constants.OUTPUTFOLDER)) {
				return p.getValue();
			}

		}
		return "";

	}

	public void setOutputFolder(String outputFolder) {
		UniqueProperty p = new UniqueProperty();
		p.setName(Constants.OUTPUTFOLDER);
		p.setValue(outputFolder);
		if (!usage.contains(p))
			usage.add(p);
		else
			for (Property property : usage)
				if (property.getName().equals(Constants.OUTPUTFOLDER))
					property.setValue(outputFolder);

	}

	public String getInvocation() {
		for (Property p : usage) {
			if (p.getName().equals(Constants.INVOCATION)) {
				return p.getValue();
			}
		}
		return "";
	}

	@Override
	public Element serialize() throws ParserConfigurationException {
		Serializer rootSer = super.serializer();
		Serializer ser = new Serializer(
				DataMiningSoftwareConstants.DataMiningSoftware);
		ser.addAsAttribute(Constants.Name, name);
		ser.addAsChildElement(Constants.HOSTNAME, this.getHostName());
		rootSer.addAsChildElement(ser.serializedElement());
		return rootSer.serializedElement();
	}

}
