/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.workflow.plugin;

import java.net.UnknownHostException;
import java.util.HashSet;
import java.util.Set;

import javax.xml.parsers.ParserConfigurationException;

import kgrid.raems.service.impl.Consts;
import kgrid.raems.service.impl.Utils;
import kgrid.raems.service.impl.utils.ID;
import kgrid.raems.service.impl.utils.SetUtils;
import kgrid.raems.service.impl.workflow.Attribute;
import kgrid.raems.service.impl.workflow.Data;
import kgrid.raems.service.impl.workflow.Deserializer;
import kgrid.raems.service.impl.workflow.PluginDescription;
import kgrid.raems.service.impl.workflow.Property;
import kgrid.raems.service.impl.workflow.Resource;
import kgrid.raems.service.impl.workflow.Serializer;
import kgrid.raems.service.impl.workflow.constants.Constants;
import kgrid.raems.service.impl.workflow.constants.DatasetConstants;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class Dataset extends Data implements Consts {
	private String hostName;
	private Set<UniqueProperty> access;
	private Set<UniqueProperty> format;
	private Set<Attribute> attributes;

	public Dataset() {
		super();
		access = new HashSet<UniqueProperty>();
		format = new HashSet<UniqueProperty>();
		attributes = new HashSet<Attribute>();
		hostName = "";
	}

	public String getPhysicalLocation() {
		for (Property p : this.access) {
			if (p.getName().equals("Location")) {
				return p.getValue();
			}
		}
		return "";
	}

	public void update(Document metadata) throws Exception {
		PluginDescription pluginDescription = new PluginDescription("Dataset");
		Deserializer deserializer = new Deserializer(pluginDescription);
		Dataset d = (Dataset) deserializer.deserialize(metadata);
		this.name = d.name;
		this.access = d.access;
		this.attributes = d.attributes;
		this.format = d.format;
		this.type = d.type;

	}

	public String toString() {
		String s = "[Dataset\n";
		if (id != null)
			s += id + "\n";
		if (href != null && !href.equals(""))
			s += "(href: " + href + ")\n";
		if (title != null && !title.equals(""))
			s += "(title: " + title + ")\n";
		if (name != null && !name.equals(""))
			s += "(name: " + name + ")\n";
		if (hostName != null && !hostName.equals(""))
			s += "(hostname: " + hostName + ")\n";
		if (type != null && !type.equals(""))
			s += "(type: " + type + ")\n";
		if (access != null && access.size() > 0) {
			s += "(access:\n";
			for (Property p : access)
				s += p.toString() + "\n";
		}
		s += ")\n";
		if (format != null && format.size() > 0) {
			s += "(format:\n";
			for (Property p : format)
				s += p.toString() + "\n";
		}
		s += ")\n";
		if (attributes != null && attributes.size() > 0) {
			s += "(attributes:\n";
			for (Attribute p : attributes)
				s += p.toString() + "\n";
		}
		return s + ")]\n";

	}

	public int getAccessPropertiesNumber() {
		return access.size();
	}

	public int getFormatPropertiesNumber() {
		return format.size();
	}

	public int getAttributesNumber() {
		return attributes.size();
	}

	public Set<UniqueProperty> getAccess() {
		return access;
	}

	public void setAccess(Set<UniqueProperty> access) {
		this.access = access;
	}

	public Set<UniqueProperty> getFormat() {
		return format;
	}

	public void setFormat(Set<UniqueProperty> format) {
		this.format = format;
	}

	public Set<Attribute> getAttributes() {
		return attributes;
	}

	public void setAttributes(Set<Attribute> attributes) {
		this.attributes = attributes;
	}

	public void addAccessProperty(UniqueProperty property) {
		access.add(property);
	}

	public void addFormatProperty(UniqueProperty property) {
		format.add(property);
	}

	public void addAttribute(Attribute attribute) {
		attributes.add(attribute);
	}

	public Resource clone() throws CloneNotSupportedException {
		Dataset ret = new Dataset();
		ret.setType(new String(this.type));
		ret.setId((ID) id.clone());
		ret.setName(new String(this.name));
		ret.setConcrete(new String(this.getHostName()),
				new String(this.getPhysicalLocation()));
		ret.setPublicInformation(new String(this.href), new String(this.title));
		try {
			Set<UniqueProperty> set1 = (Set<UniqueProperty>) SetUtils
					.cloneSet(this.access);
			ret.setAccess(set1);
			Set<UniqueProperty> set2 = (Set<UniqueProperty>) SetUtils
					.cloneSet(this.format);
			ret.setFormat(set2);
			Set<Attribute> set3 = (Set<Attribute>) SetUtils
					.cloneSet(this.attributes);
			ret.setAttributes(set3);
		} catch (Throwable t) {
			throw new CloneNotSupportedException();
		}
		return ret;
	}

	public String getHostName() {
		return hostName;
	}

	public void setConcrete(String hostname, String path) {
		this.hostName = hostname;
		UniqueProperty loc = new UniqueProperty();
		loc.setName(Constants.LOCATION);
		loc.setValue(path);
		if (!access.contains(loc)) {
			addAccessProperty(loc);
		} else {
			for (Property p : this.access) {
				if (p.getName().equals(Constants.LOCATION)) {
					p.setValue(path);

				}
			}
		}
	}

	@Override
	public Element serialize() throws ParserConfigurationException {
		Serializer ser = new Serializer(DatasetConstants.Dataset);
		ser.addAsAttribute(Constants.Name, this.name);
		ser.addAsAttribute(Constants.Type, this.type);
		return ser.serializedElement();
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Dataset) {
			Dataset d = (Dataset) obj;
			try {
				return Utils.sameHosts(hostName, d.hostName)
						&& getPhysicalLocation()
								.equals(d.getPhysicalLocation());
			} catch (UnknownHostException e) {
				e.printStackTrace();
			}
		}
		return false;
	}
}
