/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675 Mass
 * Ave, Cambridge, MA 02139, USA.
 * 
 * =============================================================================
 * Grid Computing Laboratory, Dept. of Electronics, Computer Science and
 * Systems, University of Calabria, Italy. http://grid.deis.unical.it/
 * =============================================================================
 */

package kgrid.raems.service.impl.workflow.plugin;

import kgrid.raems.service.impl.exception.ParserException;
import kgrid.raems.service.impl.workflow.Data;
import kgrid.raems.service.impl.workflow.Deserializer;
import kgrid.raems.service.impl.workflow.PluginDescription;
import kgrid.raems.service.impl.workflow.Program;
import kgrid.raems.service.impl.workflow.Resource;
import kgrid.raems.service.impl.workflow.constants.Constants;
import kgrid.raems.service.impl.workflow.parser.Builder;
import kgrid.raems.service.impl.workflow.parser.Parser;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * @author Eugenio Cesario <cesario@icar.cnr.it>
 * @author Marco Lackovic <mlackovic@deis.unical.it>
 */
public class JobParser implements Parser {

	private JobBuilder builder;

	public void parseExecutionPlan(Node node) {
		builder.buildJob();

	}

	public void parseTask(Node node) {
		NamedNodeMap attributes = node.getAttributes();
		Node label = attributes.getNamedItem(Constants.EpLabel);
		Node name = attributes.getNamedItem(Constants.Name);
		if (label != null) {
			builder.buildAbstractTask(label.getTextContent());
		} else if (name != null) {
			builder.buildConcreteTask(name.getTextContent());
		} else {
			throw new ParserException("Malformed xml");
		}
	}

	public void parseTaskLink(Node node) {
		NamedNodeMap attributes = node.getAttributes();
		if (attributes == null) {
			throw new ParserException("Malformed xml");
		}
		Node from_att = attributes.getNamedItem(Constants.EpFrom);
		Node to_att = attributes.getNamedItem(Constants.EpTo);
		if (from_att == null || to_att == null) {
			throw new ParserException("Malformed xml");
		}
		String from = from_att.getNodeValue();
		String to = to_att.getNodeValue();
		builder.buildLink(from, to);

	}

	public void parseComputation(Node node) {
		builder.buildComputation();
	}

	public void parseDataTransfer(Node node) {
		NamedNodeMap attrs = node.getAttributes();
		Node source = attrs.getNamedItem(Constants.Source);
		Node destination = attrs.getNamedItem(Constants.Destination);
		String sourceS = "", destinationS = "";
		if (source != null)
			sourceS = source.getTextContent();
		else
			throw new ParserException("Malformed xml");
		if (destination != null)
			destinationS = destination.getTextContent();
		else
			throw new ParserException("Malformed xml");
		builder.buildDataTransfer(sourceS, destinationS);
	}

	public void parseInput(Node node) throws Exception {
		NamedNodeMap attrs = node.getAttributes();
		Node idref = attrs.getNamedItem(Constants.IdRef);
		Node id = attrs.getNamedItem(Constants.Id);

		if (idref == null && id != null) {
			String classname = getFirstNonTextNode(node);
			PluginDescription pluginDescription = new PluginDescription(
					classname);
			Deserializer d = new Deserializer(pluginDescription);

			Resource i = (Resource) d.deserialize(node);
			builder.buildInputOfTask(i);
		} else {
			builder.buildInputOfTask(idref.getTextContent());
		}

	}

	public void parseOutput(Node node) throws Exception {
		String classname = getFirstNonTextNode(node);
		PluginDescription pluginDescription = new PluginDescription(classname);
		Deserializer d = new Deserializer(pluginDescription);
		Data o = (Data) d.deserialize(node);
		builder.buildOutputOfComputation(o);
	}

	private String getFirstNonTextNode(Node node) {
		NodeList l = node.getChildNodes();
		for (int i = 0; i < l.getLength(); i++) {
			Node n = l.item(i);
			if (!n.getNodeName().equals("#text")) {
				return n.getNodeName();
			}
		}
		return null;
	}

	public void parseProgram(Node node) throws Exception {
		String classname = getFirstNonTextNode(node);
		PluginDescription pluginDescription = new PluginDescription(classname);
		Deserializer d = new Deserializer(pluginDescription);
		Program p = (Program) d.deserialize(node);
		builder.buildProgramOfComputation(p);
	}

	public void parseElement(Node node) {

	}

	public Object endParsing() {
		return builder.getResult();
	}

	@Override
	public void setBuilder(Builder b) {
		this.builder = (JobBuilder) b;
	}

}
